# -*- coding: utf-8 -*-
# wasp_general/task/health.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: this idea and implementation require to be more useful. Till that this module should be treated as deprecated

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod
from enum import Enum

from wasp_general.verify import verify_type, verify_value


class WTaskHealthSensor(metaclass=ABCMeta):
	""" Represent single sensor
	"""

	class WTaskSensorSeverity(Enum):
		""" Sensor severity
		"""
		recommended = 0
		important = 1
		critical = 2

	@verify_type(name=str, severity=WTaskSensorSeverity, description=(str, None))
	def __init__(self, name, severity, description=None):
		""" Construct new sensor

		:param name: sensor name
		:param severity: sensor severity
		:param description: sensor description
		"""
		self.__name = name
		self.__severity = severity
		self.__description = description

	def name(self):
		""" Return sensor name

		:return: string
		"""
		return self.__name

	def description(self):
		""" Return sensor description

		:return: string (or None if there is no description)
		"""
		return self.__description

	def severity(self):
		""" Return sensor severity

		:return: WTaskHealthSensor.WTaskSensorSeverity
		"""
		return self.__severity

	@abstractmethod
	def healthy(self):
		""" Return True if sensor is OK, else - False

		:return: bool
		"""
		raise NotImplementedError("This method is abstract")


class WMeasurableTaskHealthSensor(WTaskHealthSensor, metaclass=ABCMeta):
	""" Represent measurable sensor
	"""

	@verify_value(comparator=lambda x: x is None or callable(x))
	def __init__(self, name, severity, error_value, description=None, comparator=None):
		""" Construct new measurable sensor.

		:param name: sensor name (as in :meth:`.WTaskHealthSensor.__init__`)
		:param severity: sensor severity (as in :meth:`.WTaskHealthSensor.__init__`)
		:param error_value: sensor limit
		:param description: sensor description (as in :meth:`.WTaskHealthSensor.__init__`)
		:param comparator: callable, that compares sensor value with error_value (if result is True, then \
		sensor threats as healthy, if False - unhealthy). If comparator is None, then default function \
		is used (function compares if sensor value are less then error_value)
		"""
		WTaskHealthSensor.__init__(self, name, severity, description=description)
		self._error_value = error_value
		self._comparator = comparator if comparator is not None else lambda x, y: x < y

	def healthy(self):
		""" Return True if sensor is OK, else - False

		:return: bool
		"""
		return self._comparator(self.value(), self._error_value)

	def error_value(self):
		""" Return error_value

		:return: (any type)
		"""
		return self._error_value

	@abstractmethod
	def value(self):
		""" Return current value sensor

		:return: (any type)
		"""
		raise NotImplementedError("Value wasn't defined")

	def measurement_unit(self):
		""" Return measurement unit of sensor

		:return: string
		"""
		raise NotImplementedError("Measurement wasn't defined")


class WTaskHealth(metaclass=ABCMeta):
	""" This class presents API for task health probe. It can be used to probe if everything that task needs are
	ready or available (for example: disk space, file existing, socket is opened and so on).

	This is optional for the most tasks.

	"""

	@verify_type(sensors=WTaskHealthSensor)
	def __init__(self, *sensors, decorate_start=True, decorate_stop=True):

		self._sensors = {}
		for sensor in sensors:
			# noinspection PyUnresolvedReferences
			self._sensors[sensor.name()] = sensor

	def sensor(self, sensor_name):
		""" Return sensor by its name

		:param sensor_name: name of sensor
		:return: WMeasurableTaskHealthSensor instance
		"""
		return self._sensors[sensor_name]

	def sensors(self):
		""" Return list of sensor names

		:return: list of strings
		"""
		return list(self._sensors.keys())

	def healthy(self):
		""" Return task health. If None - task is healthy, otherwise - maximum severity of sensors

		:return: None or WTaskHealthSensor.WTaskSensorSeverity
		"""
		state = None
		for sensor in self._sensors.values():
			if sensor.healthy() is False:
				if state is None or sensor.severity().value > state.value:
					state = sensor.severity()
				if state == WTaskHealthSensor.WTaskSensorSeverity.critical:
					break

		return state



# -*- coding: utf-8 -*-
# wasp_general/types/binarray.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from wasp_general.verify import verify_type, verify_value


class WBinArray:
	""" This type represent sequence of bit. Bits are ordered as MSB does (most significant bit has index 0).
	If size is specified, then array became fixed-length array. All actions beyond the limit will be omitted or
	exception will be raised.

	see also https://en.wikipedia.org/wiki/Bit_numbering
	"""

	@verify_type(size=(int, None))
	@verify_value(value=lambda x: int(x) >= 0, size=lambda x: x is None or x >= 0)
	def __init__(self, value=0, size=None):
		""" Create new bit sequence.

		:param value: value with which this sequence must be initialized.
		:param size: sequence fixed size (number of bits).
		"""
		if isinstance(value, (int, WBinArray)) is False:
			raise TypeError('Invalid array type')
		self.__value = int(value)
		self.__size = size
		if self.__size is not None:
			if self.__value > ((2 ** self.__size) - 1):
				raise ValueError('Value is out of bound')

	def __len__(self):
		""" Return size if array has fixed size, otherwise return bits count

		:return: int
		"""
		if self.__size is not None:
			return self.__size
		elif self.__value == 0:
			return 1
		else:
			return self.__value.bit_length()

	@verify_type(item=(int, slice))
	def __getitem__(self, item):
		""" Return bit value for specified index or array part defined by the slice

		:param item: index or slice object
		:return: int for single item, WBinArray for slice
		"""
		if isinstance(item, int):
			if item >= self.__len__() or item < 0:
				raise IndexError('Index out of range')
			mask = 1 << (len(self) - item - 1)
			return 1 if (self.__value & mask) > 0 else 0
		else:
			# isinstance(item, slice):
			start = item.start if item.start is not None else 0
			if start > len(self):
				raise IndexError('Index out of range')

			stop = item.stop if item.stop is not None else len(self)
			stop = stop if stop <= len(self) else len(self)

			step = item.step if item.step is not None else 1

			iterator = range(start, stop, step)
			result = WBinArray(0, len(iterator))
			for key in iterator:
				result[key - start] = self[key]

			return result

	@verify_type(key=int)
	@verify_value(value=lambda x: int(x) >= 0)
	def __setitem__(self, key, value):
		""" Set value for the given index. If value is greater then 1 or value is a WBinArray with length
		greater then 1, then all the bits are copied from original value to this array starting by the key
		index.

		:param key: starting position
		:param value: value to set
		:return: None
		"""

		if isinstance(value, (int, WBinArray)) is False:
			raise TypeError('Invalid type for value')
		value = int(value)

		if self.__size is not None:
			if key < 0 or key >= len(self):
				raise IndexError('Index out of range')
			length = len(WBinArray(value))
			if (key + length - 1) >= len(self):
				raise OverflowError('Value is bigger then array')

		if value == 0:
			mask = ~(1 << (len(self) - key - 1))
			self.__value = (self.__value & mask)
		else:
			# value > 0
			mask = (1 << (len(self) - key - 1))
			self.__value = (self.__value | mask)

	def __int__(self):
		""" Convert to integer

		:return: int
		"""
		if self.__size is None:
			return self.__value
		else:
			mask = ((2 ** self.__size) - 1)
			return self.__value & mask

	def __str__(self):
		""" Convert to string

		:return: str
		"""
		format_str = "{:0%ib}" % len(self)
		return format_str.format(self.__int__())

	def __add__(self, other):
		""" Return addition of two objects

		:param other: second summand
		:return: int
		"""
		return self.__int__() + other.__int__()

	def __sub__(self, other):
		""" Return subtraction of two objects

		:param other: subtrahend
		:return: int
		"""
		return self.__int__() - other.__int__()

	@verify_type(size=(int, None))
	@verify_value(size=lambda x: x >= 0)
	def resize(self, size):
		""" Resize current array. If size is None, then array became nonfixed-length array. If new size is
		less then current size and value, then value will be truncated (lesser significant bits will be
		truncated).

		:param size:
		:return:
		"""
		if size is not None:
			self.__value = int(WBinArray(self.__value)[:size])
		self.__size = size

	@verify_type(bits_count=int)
	@verify_value(bits_count=lambda x: x > 0)
	def split(self, bits_count):
		""" Split array into smaller parts. Each small array is fixed-length WBinArray (length of that array is
		bits_count).

		:param bits_count: array length
		:return: list of WBinArray
		"""
		result = []
		array = WBinArray(self.__value, self.__size)

		if (len(array) % bits_count) > 0:
			array.resize(len(array) + (bits_count - (len(array) % bits_count)))

		while len(array):
			result.append(WBinArray(array[:bits_count], bits_count))
			array = array[bits_count:]
		return result

	def concat(self, array):
		""" Return new fixed-length array, that is made by creating new array with length of sum of two arrays
		(this array and the given one). In newly created array the most significant bit of the given array
		will have an index lesser then an index of the least significant bit of this array.

		:param array: array to concatenate with
		:return: WBinArray
		"""
		if isinstance(array, WBinArray) is False:
			raise TypeError('Invalid array type')
		value = ((int(self) << len(array)) | int(array))
		return WBinArray(value, len(self) + len(array))

	def rconcat(self, array):
		""" 'Reverse' concatenation. Works the same as :meth:`.WBinArray.concat`, but in newly created array
		the most significant bit of the given array will have an index greater then an index of the least
		significant bit of this array

		:param array: array to concatenate with
		:return: WBinArray
		"""
		if isinstance(array, WBinArray) is False:
			raise TypeError('Invalid array type')
		return array.concat(self)

	def extend(self, *array_list):
		""" Concatenate this array with the given arrays. This method doesn't modify current array. Instead,
		it creates new one, that have all of arrays. (see :meth:`.WBinArray.concat` method)

		:param array_list: list of WBinArray
		:return: newly created WBinArray
		"""
		result = WBinArray(int(self), len(self))
		for array in array_list:
			result = result.concat(array)
		return result

	def swipe(self):
		""" Mirror current array value in reverse. Bits that had greater index will have lesser index, and
		vice-versa. This method doesn't change this array. It creates a new one and return it as a result.

		:return: WBinArray
		"""
		result = WBinArray(0, len(self))
		for i in range(len(self)):
			result[len(self) - i - 1] = self[i]
		return result

	@staticmethod
	def join(*args):
		""" Concatenate all of the given arrays. (see :meth:`.WBinArray.concat` method)

		:param args: list of WBinArray
		:return: WBinArray
		"""
		if len(args) == 0:
			return

		result = WBinArray(int(args[0]), len(args[0]))
		if len(args) == 1:
			return result
		else:
			result = result.extend(*(args[1:]))
		return result



# -*- coding: utf-8 -*-
# wasp_general/network/web/response.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from wasp_general.verify import verify_type, verify_value

from wasp_general.network.web.proto import WWebResponseProto
from wasp_general.network.web.headers import WHTTPHeaders


class WWebResponse(WWebResponseProto):
	""" Simple :class:`.WWebResponseProto` implementation
	"""

	@verify_type(status=(int, None), headers=(WHTTPHeaders, None), response_data=(bytes, None))
	@verify_value(status=lambda x: x is None or x > 0)
	def __init__(self, status=None, headers=None, response_data=None):
		""" Create new response

		:param status: response status code
		:param headers: response headers
		:param response_data: response data
		"""
		WWebResponseProto.__init__(self)
		self.__status = status
		self.__headers = headers
		self.__response_data = response_data
		self.__pushed_responses = []

	def status(self):
		""" :meth:`.WWebResponseProto.status` method implementation
		"""
		return self.__status

	def headers(self):
		""" :meth:`.WWebResponseProto.headers` method implementation
		"""
		return self.__headers

	def response_data(self):
		""" :meth:`.WWebResponseProto.response_data` method implementation
		"""
		return self.__response_data

	@verify_type(response=WWebResponseProto)
	def __push__(self, *responses):
		""" Save responses to push

		:param responses: responses to push
		:return:
		"""
		self.__pushed_responses.extend(responses)

	def __pushed_responses__(self):
		""" :meth:`.WWebResponseProto.__pushed_responses__` method implementation
		"""
		return tuple(self.__pushed_responses)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.network.beacon.messenger import WBeaconMessengerBase, WBeaconMessenger, WBeaconGouverneurMessenger
from wasp_general.network.beacon.messenger import WHostgroupBeaconMessenger
from wasp_general.network.primitives import WIPV4SocketInfo
from wasp_general.config import WConfig


def test_abstract():
	pytest.raises(TypeError, WBeaconMessengerBase)
	pytest.raises(NotImplementedError, WBeaconMessengerBase.request, None, WConfig())
	pytest.raises(NotImplementedError, WBeaconMessengerBase.has_response, None, WConfig(), b'', WIPV4SocketInfo())
	pytest.raises(NotImplementedError, WBeaconMessengerBase.response, None, WConfig(), b'', WIPV4SocketInfo())
	pytest.raises(
		NotImplementedError, WBeaconMessengerBase.response_address, None, WConfig(), b'', WIPV4SocketInfo()
	)
	pytest.raises(
		NotImplementedError, WBeaconMessengerBase.valid_response, None, WConfig(), b'', WIPV4SocketInfo()
	)


class TestWBeaconMessenger:

	def test_messenger(self):
		config = WConfig()
		client1 = WIPV4SocketInfo()

		m = WBeaconMessenger()
		assert(m.request(config) == WBeaconMessenger.__beacon_hello_msg__)
		assert(m.response(config, b'zxc', client1) == b'zxc')
		assert(m.response(config, b'1qaz', client1) == b'1qaz')

		assert(m.has_response(config, b'', client1) is False)
		assert(m.has_response(config, b'zxc', client1) is True)
		assert(m.has_response(config, b'1qaz', client1) is True)

		client2 = WIPV4SocketInfo('127.0.0.0', 200)
		assert(m.response_address(config, b'', client1).pair() == ('', 0))
		assert(m.response_address(config, b'', client2).pair() == ('127.0.0.0', 200))

		assert(m.valid_response(config, b'', client1) is False)
		assert(m.valid_response(config, b'zxc', client1) is True)


class TestWBeaconGouverneurMessenger:

	def test_request(self):
		config = WConfig()
		config.add_section('wasp-general::network::beacon')
		config['wasp-general::network::beacon']['public_address'] = ''
		config['wasp-general::network::beacon']['public_port'] = ''

		assert(WBeaconGouverneurMessenger(b'hi').request(config) == b'hi')
		assert(WBeaconGouverneurMessenger(b'hello_string').request(config) == b'hello_string')
		config['wasp-general::network::beacon']['public_address'] = 'qqq'
		assert (WBeaconGouverneurMessenger(b'hi').request(config) == b'hi:qqq')
		config['wasp-general::network::beacon']['public_address'] = ':'
		pytest.raises(ValueError, WBeaconGouverneurMessenger(b'hi').request, config)
		config['wasp-general::network::beacon']['public_address'] = 'qqq'
		config['wasp-general::network::beacon']['public_port'] = '70'
		assert(WBeaconGouverneurMessenger(b'hi').request(config) == b'hi:qqq:70')
		config['wasp-general::network::beacon']['public_address'] = ''
		assert(WBeaconGouverneurMessenger(b'hi').request(config) == b'hi')

	def test_invert_hello(self):
		assert(WBeaconGouverneurMessenger(b'hi').invert_hello() is False)
		assert(WBeaconGouverneurMessenger(b'hi', invert_hello=True).invert_hello() is True)

	def test_hello_message(self):
		assert(WBeaconGouverneurMessenger(b'hi').hello_message() == b'hi')
		assert(WBeaconGouverneurMessenger(b'hi').hello_message(invert_hello=True) == b'ih')

	def test_has_response(self):
		config = WConfig()
		m1 = WBeaconGouverneurMessenger(b'hi')
		m2 = WBeaconGouverneurMessenger(b'hello_string')

		assert(m1.has_response(config, b'hi', WIPV4SocketInfo()) is True)
		assert(m1.has_response(config, b'ih', WIPV4SocketInfo()) is False)
		assert(m2.has_response(config, b'hi', WIPV4SocketInfo()) is False)
		assert(m2.has_response(config, b'hello_string', WIPV4SocketInfo()) is True)
		assert(m1.has_response(config, b'hi:foo:1', WIPV4SocketInfo()) is True)

	def test_response(self):
		config = WConfig()
		config.add_section('wasp-general::network::beacon')
		config['wasp-general::network::beacon']['public_address'] = ''
		config['wasp-general::network::beacon']['public_port'] = ''
		si = WIPV4SocketInfo('192.168.0.1', 7777)

		assert(WBeaconGouverneurMessenger(b'hi').response(config, b'req', si) == b'hi')
		assert(WBeaconGouverneurMessenger(b'hello_string').response(config, b'req', si) == b'hello_string')
		config['wasp-general::network::beacon']['public_address'] = 'qqq'
		assert (WBeaconGouverneurMessenger(b'hi').response(config, b'req', si) == b'hi:qqq')
		config['wasp-general::network::beacon']['public_address'] = ':'
		pytest.raises(ValueError, WBeaconGouverneurMessenger(b'hi').response, config, b'req', si)
		config['wasp-general::network::beacon']['public_address'] = 'qqq'
		config['wasp-general::network::beacon']['public_port'] = '70'
		assert(WBeaconGouverneurMessenger(b'hi').response(config, b'req', si) == b'hi:qqq:70')
		config['wasp-general::network::beacon']['public_address'] = ''
		assert(WBeaconGouverneurMessenger(b'hi').response(config, b'req', si) == b'hi')

	def test_response_address(self):
		config = WConfig()
		si = WIPV4SocketInfo('192.168.0.1', 7777)

		address = WBeaconGouverneurMessenger(b'hi').response_address(config, b'hi', si)
		assert(address.pair() == ('192.168.0.1', 7777))
		address = WBeaconGouverneurMessenger(b'hi').response_address(config, b'hi:127.0.0.1', si)
		assert(address.pair() == ('127.0.0.1', 7777))
		address = WBeaconGouverneurMessenger(b'hi').response_address(config, b'hi:127.0.0.1:90', si)
		assert(address.pair() == ('127.0.0.1', 90))

		pytest.raises(ValueError, WBeaconGouverneurMessenger(b'hi').response_address, config, b'hello', si)
		pytest.raises(ValueError, WBeaconGouverneurMessenger(b'hi').response_address, config, b'hi:e:l:l:o', si)

	def test_valid_response(self):
		config = WConfig()
		m = WBeaconGouverneurMessenger(b'hi')

		assert(m.valid_response(config, b'hi', WIPV4SocketInfo()) is True)
		assert(m.valid_response(config, b'ih', WIPV4SocketInfo()) is False)

		m = WBeaconGouverneurMessenger(b'hi', invert_hello=True)
		assert(m.valid_response(config, b'hi', WIPV4SocketInfo()) is False)
		assert(m.valid_response(config, b'ih', WIPV4SocketInfo()) is True)


class TestWHostgroupBeaconMessenger:

	def test_task(self):
		config = WConfig()
		config.add_section('wasp-general::network::beacon')
		config['wasp-general::network::beacon']['public_address'] = ''
		config['wasp-general::network::beacon']['public_port'] = ''
		si = WIPV4SocketInfo('1.1.1.1', 1)

		pytest.raises(TypeError, WHostgroupBeaconMessenger, b'HELLOMSG', b'group')
		pytest.raises(ValueError, WHostgroupBeaconMessenger, b'HELLOMSG', 'group-')

		messenger = WHostgroupBeaconMessenger(b'HELLOMSG')
		assert(isinstance(messenger, WBeaconGouverneurMessenger) is True)
		assert(messenger.hostgroups() == [])
		assert(messenger.request(config) == b'HELLOMSG')
		config['wasp-general::network::beacon']['public_address'] = 'site1.org'
		assert(messenger.request(config) == b'HELLOMSG:site1.org')
		config['wasp-general::network::beacon']['public_port'] = '10'
		assert(messenger.request(config) == b'HELLOMSG:site1.org:10')
		data = messenger._message_hostgroup_parse(b'HELLOMSG:site1.org')
		assert(data[0] == [])
		assert(str(data[1].address()) == 'site1.org')
		assert(data[1].port() is None)

		assert(messenger.has_response(config, b'HELLOMSG', si) is True)
		assert(messenger.has_response(config, b'HELLOMSG:2.2.2.2', si) is True)
		assert(messenger.has_response(config, b'HELLOMSG:2.2.2.2#server', si) is True)
		assert(messenger.has_response(config, b'HELLOMSG#users', si) is True)
		assert(messenger.has_response(config, b'HELLOMSG#g', si) is True)

		config['wasp-general::network::beacon']['public_address'] = ''
		config['wasp-general::network::beacon']['public_port'] = ''
		messenger = WHostgroupBeaconMessenger(b'HELLOMSG', 'group1', 'servers')
		assert(messenger.hostgroups() == ['group1', 'servers'])
		assert(messenger.request(config) == b'HELLOMSG#group1,servers')
		config['wasp-general::network::beacon']['public_address'] = 'site1.org'
		assert(messenger.request(config) == b'HELLOMSG:site1.org#group1,servers')
		data = messenger._message_hostgroup_parse(b'HELLOMSG:site1.org#group1,servers')
		assert(data[0] == [b'group1', b'servers'])
		assert(str(data[1].address()) == 'site1.org')
		assert(data[1].port() is None)
		config['wasp-general::network::beacon']['public_port'] = '10'
		assert(messenger.request(config) == b'HELLOMSG:site1.org:10#group1,servers')

		pytest.raises(ValueError, messenger._message_hostgroup_parse, b'HELLOMSG##group1,servers')

		assert(messenger.has_response(config, b'HELLOMSG', si) is True)
		assert(messenger.has_response(config, b'HELLOMSG:2.2.2.2', si) is True)
		assert(messenger.has_response(config, b'HELLOMSG:2.2.2.2#servers', si) is True)
		assert(messenger.has_response(config, b'HELLOMSG:2.2.2.2#servers, group1', si) is True)
		assert(messenger.has_response(config, b'HELLOMSG#users', si) is False)
		assert(messenger.has_response(config, b'HELLOMSG#g', si) is False)
		assert(messenger.has_response(config, b'HELLOMSG##servers', si) is False)

		assert(str(messenger.response_address(config, b'HELLOMSG', si).address()) == '1.1.1.1')
		assert(str(messenger.response_address(config, b'HELLOMSG:2.2.2.2', si).address()) == '2.2.2.2')
		assert(str(messenger.response_address(config, b'HELLOMSG#users', si).address()) == '1.1.1.1')
		assert(str(messenger.response_address(config, b'HELLOMSG:2.2.2.2#u', si).address()) == '2.2.2.2')



# -*- coding: utf-8 -*-
# wasp_general/command/context.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: tests require

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type, verify_value
from wasp_general.command.command import WCommandProto
from wasp_general.composer import WComposerProto


class WContextProto(metaclass=ABCMeta):
	""" Represent context configuration
	"""

	@abstractmethod
	def context_name(self):
		""" Return this context name

		:return: str
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def context_value(self):
		""" Return this context value (can be None)

		:return: str or None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def linked_context(self):
		""" Return link to 'parent'/'higher' context

		:return: WContextProto or None
		"""
		raise NotImplementedError('This method is abstract')

	def __len__(self):
		""" Return linked context count

		:return: int
		"""
		return len([x for x in self])

	def __iter__(self):
		"""Iterate over

		:return:
		"""
		context = self
		while context is not None:
			yield context
			context = context.linked_context()

	def __eq__(self, other):
		""" Compare two context. Two context are equal if they have the same context_name and each their own
		linked context are equal also.

		:param other: context to compare
		:return: bool
		"""
		if isinstance(other, WContextProto) is False:
			return False

		context_a = self
		context_b = other

		while context_a is not None and context_b is not None:
			if context_a.context_name() != context_b.context_name():
				return False

			context_a = context_a.linked_context()
			context_b = context_b.linked_context()

		if context_b is not None:
			return False
		elif context_a is not None:
			return False
		return True


class WContext(WContextProto):
	""" :class:`.WContextProto` implementation
	"""

	@verify_type(context_name=str, context_value=(str, None), linked_context=(WContextProto, None))
	def __init__(self, context_name, context_value=None, linked_context=None):
		""" Create new context request

		:param context_name: context name
		:param context_value: context value
		:param linked_context: linked context
		"""
		self.__context_name = context_name
		self.__context_value = context_value
		self.__linked_context = linked_context

	def context_name(self):
		""" :meth:`.WContextProto.context_name` implementation
		"""
		return self.__context_name

	def context_value(self):
		""" :meth:`.WContextProto.context_value` implementation
		"""
		return self.__context_value

	def linked_context(self):
		""" :meth:`.WContextProto.linked_context` implementation
		"""
		return self.__linked_context

	@classmethod
	@verify_type(context=(WContextProto, None))
	def export_context(cls, context):
		""" Export the specified context to be capable context transferring

		:param context: context to export
		:return: tuple
		"""
		if context is None:
			return
		result = [(x.context_name(), x.context_value()) for x in context]
		result.reverse()
		return tuple(result)

	@classmethod
	@verify_type(context=(tuple, list, None))
	def import_context(cls, context):
		""" Import context to corresponding WContextProto object (:meth:`WContext.export_context` reverse operation)

		:param context: context to import
		:return: WContext
		"""
		if context is None or len(context) == 0:
			return

		result = WContext(context[0][0], context[0][1])
		for iter_context in context[1:]:
			result = WContext(iter_context[0], context_value=iter_context[1], linked_context=result)
		return result

	@classmethod
	@verify_type(context_specs=str)
	def specification(cls, *context_specs):
		""" Return linked context as adapter specification (is used by :class:`.WCommandContextAdapter`)

		:param context_specs: context names
		:return: WContext
		"""
		import_data = []
		for name in context_specs:
			import_data.append((name, None))
		return cls.import_context(import_data)


class WContextComposer(WComposerProto):

	@verify_type('paranoid', obj_spec=(tuple, list, None))
	def compose(self, obj_spec):
		return WContext.import_context(obj_spec)

	@verify_type('paranoid', obj=(WContextProto, None))
	def decompose(self, obj):
		return WContext.export_context(obj)


class WCommandContextAdapter(metaclass=ABCMeta):
	""" Adapter is used for command tokens modification
	"""

	@verify_type(context_specifications=(WContextProto, None))
	def __init__(self, context_specifications):
		""" Create adapter

		:param context_specifications: context for what this adapter works
		"""
		self.__spec = context_specifications

	def specification(self):
		""" Return adapter specification

		:return: WContextProto or None
		"""
		return self.__spec

	@verify_type(command_context=(WContextProto, None))
	def match(self, command_context=None, **command_env):
		""" Check if context request is compatible with adapters specification. True - if compatible,
		False - otherwise

		:param command_context: context to check
		:param command_env: command environment
		:return: bool
		"""
		spec = self.specification()
		if command_context is None and spec is None:
			return True
		elif command_context is not None and spec is not None:
			return command_context == spec
		return False

	@abstractmethod
	@verify_type(command_tokens=str, command_context=(WContextProto, None))
	def adapt(self, *command_tokens, command_context=None, **command_env):
		""" Adapt the given command tokens with this adapter

		:param command_tokens: command tokens to adapt
		:param command_context: context
		:param command_env: command environment
		:return: list of str
		"""
		raise NotImplementedError('This method is abstract')


class WCommandContext(WCommandProto):
	""" Command that can be adapted by a context
	"""

	@verify_type(command=WCommandProto, context_adapter=WCommandContextAdapter)
	def __init__(self, base_command, context_adapter):
		""" Create new command

		:param base_command: basic command that does real magic
		:param context_adapter: adapter for command tokens modification
		"""
		WCommandProto.__init__(self)
		self.__command = base_command
		self.__adapter = context_adapter

	def original_command(self):
		""" Return source command

		:return: WCommandProto
		"""
		return self.__command

	def adapter(self):
		""" Return command adapter

		:return: WCommandAdapter
		"""
		return self.__adapter

	@verify_type('paranoid', command_tokens=str, command_context=(WContextProto, None))
	def match(self, *command_tokens, command_context=None, **command_env):
		""" Match command

		:param command_tokens: command tokens to check
		:param command_context: command context
		:param command_env: command environment
		:return: bool
		"""
		if self.adapter().match(command_context, **command_env) is False:
			return False

		command_tokens = self.adapter().adapt(*command_tokens, command_context=command_context, **command_env)
		return self.original_command().match(*command_tokens, command_context=command_context, **command_env)

	@verify_type('paranoid', command_tokens=str, command_context=(WContextProto, None))
	def exec(self, *command_tokens, command_context=None, **command_env):
		""" Execute command

		:param command_tokens: command tokens to execute
		:param command_context: command context
		:param command_env: command environment
		:return: WCommandResultProto
		"""
		if self.adapter().match(command_context, **command_env) is False:
			cmd = WCommandProto.join_tokens(*command_tokens)
			spec = self.adapter().specification()
			if spec is not None:
				spec = [x.context_name() for x in spec]
				spec.reverse()
				spec = ','.join(spec)
			raise RuntimeError('Command mismatch: %s (context: %s)' % (cmd, spec))

		command_tokens = self.adapter().adapt(*command_tokens, command_context=command_context, **command_env)
		return self.original_command().exec(*command_tokens, command_context=command_context, **command_env)



# -*- coding: utf-8 -*-
# wasp_general/network/web/re_statements.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__


http_header_name = '[^\x00-\x1e\x7f()<>@,;:\\\\"/\[\]?={} \t]+'.encode('us-ascii')
"""
Represent valid HTTP header name.

see RFC 1945, Section 4.2
external usage: :class:`.WHTTPHeaders`
"""

http_method_name = 'GET|HEAD|POST|PUT|DELETE|CONNECT|OPTIONS|TRACE'
"""
Represent valid HTTP method. This regexp has as strict restriction as HTTP/2 (RFC 7231) does.

see RFC 1945, Section 5.1.1. RFC 2616, Section 5.1.1, RFC 7231, Section 4.1

external usage: :class:`.WWebRequest`
"""

http_path_alphabet = 'a-zA-Z0-9_\-.~%:/?#\[\]@!$&\'()*+,;=\"'
"""
Represent valid set of chars, that URI can have

for URI see RFC 3986, Section 3
"""

http_path = '/[' + http_path_alphabet + ']*'
"""
Represent valid requested URI

for URI see RFC 3986, Section 3
external usage: :class:`.WWebRequest`, :class:`.WHTTPCookie`
"""

http_version = '0\.9|1\.0|1\.1'
"""
Represent valid requested HTTP-versions

see RFC 2616, Section 3.1
"""

uri_query_alphabet = 'a-zA-Z0-9\-._~%!$&\'()*+,;=:@/?'
"""
Represent valid set of chars, that query can have it in URI

see RFC 2616, Section 3.1
"""

uri_fragment_alphabet = 'a-zA-Z0-9\-._~%!$&\'()*+,;=:@/?'
"""
Represent valid set of chars, that fragment can have it in URI

see RFC 2616, Section 3.1
"""

http_get_vars_selection = '\?([' + uri_query_alphabet + ']+)(#[' + uri_fragment_alphabet + ']*)?$'
"""
Allow to select variables (GET-variables) and fragment code from URI

see RFC 2616, Section 3.4
external usage: :class:`.WWebRequestProto`
"""

http_post_vars_selection = '([' + uri_query_alphabet + ']+)'
"""
Allow to select variables that are given as 'application/x-www-form-urlencoded'.

see RFC 1867
external usage: :class:`.WWebRequestProto`
"""

http_cookie_expires = '[a-zA-Z0-9 ,:-]+'
"""
Represent valid Expires attribute value of cookie

see RFC 6265, Section 5.2.1
external usage: :class:`.WHTTPCookie`
"""

http_cookie_max_age = '[1-9][0-9]*'
"""
Represent valid Max-Age attribute value of cookie

see RFC 6265, Section 5.2.2
external usage: :class:`.WHTTPCookie`
"""

http_cookie_domain = '[a-zA-Z\-.0-9]+'
"""
Represent valid Domain attribute value of cookie

see RFC 6265, Section 5.2.3
external usage: :class:`.WHTTPCookie`
"""

http_cookie_secure = '.*'
"""
Represent valid Secure attribute value of cookie

see RFC 6265, Section 5.2.5
external usage: :class:`.WHTTPCookie`
"""

http_cookie_httponly = '.*'
"""
Represent valid HttpOnly attribute value of cookie

see RFC 6265, Section 5.2.6
external usage: :class:`.WHTTPCookie`
"""

wasp_presenter_name_alphabet = 'a-zA-Z0-9_.\-'
"""
Represent valid set of chars, that presenter name can have
"""

wasp_presenter_name_selection = '([a-zA-Z][' + wasp_presenter_name_alphabet + ']*)'
"""
Allow to select presenter name
"""

wasp_route_uri_pattern_alphabet = http_path_alphabet + '\{\}\\\\'
"""
Represent valid set of chars, that route URI pattern can have
"""

wasp_route_arg_name = '[a-zA-Z][a-zA-Z0-9_]*'
"""
Represent valid name for route argument name

external usage: :class:`.WWebRouteMap`
"""
wasp_route_arg_name_selection = '\{(' + wasp_route_arg_name + ')\}'
"""
Represent simple route argument pattern

external usage: :class:`.WWebRoute.BasicArgSearch`
"""

wasp_route_arg_value_alphabet = http_path_alphabet
for c in ['/', '?', ',', '#']:
	wasp_route_arg_value_alphabet = wasp_route_arg_value_alphabet.replace(c, '')
""" Represent valid set of chars, that route argument value can have
"""

wasp_route_arg_value_selection = '([' + wasp_route_arg_value_alphabet + ']+)'
"""
Allow to select route argument from a requested URI

external usage: :class:`.WWebRoute.BasicArgSearch`
"""

wasp_route_custom_arg_value_pattern = '[^\"]'
"""
Represent custom route argument pattern
"""

wasp_route_custom_arg_selection = \
	'(\{(' + wasp_route_arg_name + ') *: *\"(' + wasp_route_custom_arg_value_pattern + '+)\"\})'
"""
Allow to select custom route argument name and argument pattern from URI pattern

external usage: :class:`.WWebRoute.CustomArgSearch`
"""

wasp_route_import_pattern = \
	'^\s*([' + wasp_route_uri_pattern_alphabet + ']+) +=> +' + wasp_presenter_name_selection + '( +\((.*)\))?\s*$'
"""
Allow to parse custom route from a imported text

external usage: :class:`.WWebRouteMap`
"""



# -*- coding: utf-8 -*-
# wasp_general/network/messenger/coders.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from base64 import b64encode, b64decode
from enum import Enum

from wasp_general.verify import verify_type
from wasp_general.crypto.hex import WHex, WUnHex
from wasp_general.crypto.aes import WAES
from wasp_general.crypto.rsa import WRSA

from wasp_general.network.messenger.proto import WMessengerOnionSessionProto
from wasp_general.network.messenger.envelope import WMessengerTextEnvelope, WMessengerBytesEnvelope
from wasp_general.network.messenger.layers import WMessengerOnionCoderLayerProto


class WMessengerFixedModificationLayer(WMessengerOnionCoderLayerProto):
	""" :class:`.WMessengerOnionCoderLayerProto` class implementation. This class applies fixed modification to
	specified messages.

	In :meth:`.WMessengerFixedModificationLayer.encode` method this class appends "header" to the message start
	or appends "tail" to the message end. One of them must be specified.

	For :class:`.WMessengerTextEnvelope` envelope, "header" or "tail" must be "str" type.
	For :class:`.WMessengerBytesEnvelope` envelope, "header" or "tail" must be "bytes" type.
	"""

	__layer_name__ = "com.binblob.wasp-general.fixed-modification-layer"
	""" Layer name
	"""

	class Target(Enum):
		""" Modification mode. Specifies whether modification code must be appended to start or to the end
		of a message
		"""
		head = 1
		""" Modification code must be inserted to a message start
		"""
		tail = 2
		""" Modification code must be added to a message end
		"""

	def __init__(self):
		""" Construct new layer
		"""
		WMessengerOnionCoderLayerProto.__init__(self, WMessengerFixedModificationLayer.__layer_name__)

	@verify_type(envelope=(WMessengerTextEnvelope, WMessengerBytesEnvelope))
	def __args_check(self, envelope, target, modification_code):
		""" Method checks arguments, that are specified to the
		:meth:`.WMessengerFixedModificationLayer.encode` and :meth:`.WMessengerFixedModificationLayer.decode`
		methods

		:param envelope: same as envelope in :meth:`.WMessengerFixedModificationLayer.encode` and \
		:meth:`.WMessengerFixedModificationLayer.decode` methods
		:param target: same as target in :meth:`.WMessengerFixedModificationLayer.encode` and \
		:meth:`.WMessengerFixedModificationLayer.decode` methods
		:param modification_code: same as modification_code in \
		:meth:`.WMessengerFixedModificationLayer.encode` and :meth:`.WMessengerFixedModificationLayer.decode` \
		methods

		:return: None
		"""
		if target is None:
			raise RuntimeError('"target" argument must be specified for this layer')
		if modification_code is None:
			raise RuntimeError('"modification_code" argument must be specified for this layer')

		if isinstance(target, WMessengerFixedModificationLayer.Target) is False:
			raise TypeError('Invalid "target" argument')

		if isinstance(envelope, WMessengerTextEnvelope) is True:
			if isinstance(modification_code, str) is False:
				raise TypeError('Invalid "modification_code" argument for specified envelope')
		elif isinstance(modification_code, bytes) is False:
			raise TypeError('Invalid "modification_code" argument for specified envelope')

	@verify_type('paranoid', envelope=(WMessengerTextEnvelope, WMessengerBytesEnvelope))
	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	def encode(self, envelope, session, target=None, modification_code=None, **kwargs):
		""" Methods appends 'modification_code' to the specified envelope.

		:param envelope: original envelope
		:param session: original session
		:param target: flag, that specifies whether code must be appended to the start or to the end
		:param modification_code: code to append
		:param kwargs: additional arguments

		:return: WMessengerTextEnvelope or WMessengerBytesEnvelope (depends on the original envelope)
		"""
		self.__args_check(envelope, target, modification_code)

		if isinstance(envelope, WMessengerTextEnvelope):
			target_envelope_cls = WMessengerTextEnvelope
		else:  # isinstance(envelope, WMessengerBytesEnvelope)
			target_envelope_cls = WMessengerBytesEnvelope

		if target == WMessengerFixedModificationLayer.Target.head:
			return target_envelope_cls(modification_code + envelope.message(), meta=envelope)
		else:  # target == WMessengerFixedModificationLayer.Target.tail
			return target_envelope_cls(envelope.message() + modification_code, meta=envelope)

	@verify_type('paranoid', envelope=(WMessengerTextEnvelope, WMessengerBytesEnvelope))
	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	def decode(self, envelope, session, target=None, modification_code=None, **kwargs):
		""" Methods checks envelope for 'modification_code' existence and removes it.

		:param envelope: original envelope
		:param session: original session
		:param target: flag, that specifies whether code must be searched and removed at the start or at the end
		:param modification_code: code to search/remove
		:param kwargs: additional arguments

		:return: WMessengerTextEnvelope or WMessengerBytesEnvelope (depends on the original envelope)
		"""
		self.__args_check(envelope, target, modification_code)

		message = envelope.message()
		if len(message) < len(modification_code):
			raise ValueError('Invalid message length')

		if isinstance(envelope, WMessengerTextEnvelope):
			target_envelope_cls = WMessengerTextEnvelope
		else:  # isinstance(envelope, WMessengerBytesEnvelope)
			target_envelope_cls = WMessengerBytesEnvelope

		if target == WMessengerFixedModificationLayer.Target.head:
			if message[:len(modification_code)] != modification_code:
				raise ValueError('Invalid header in message')
			return target_envelope_cls(message[len(modification_code):], meta=envelope)
		else:  # target == WMessengerFixedModificationLayer.Target.tail
			if message[-len(modification_code):] != modification_code:
				raise ValueError('Invalid tail in message')
			return target_envelope_cls(message[:-len(modification_code)], meta=envelope)


class WMessengerEncodingLayer(WMessengerOnionCoderLayerProto):
	""" This layer can encode str-object to the related encoding (to the bytes-object). Or decode bytes-object from
	the specified encoding (from bytes-object to str-object)
	"""

	__layer_name__ = "com.binblob.wasp-general.encoding-layer"
	""" Layer name
	"""

	def __init__(self):
		""" Construct new layer
		"""
		WMessengerOnionCoderLayerProto.__init__(self, WMessengerEncodingLayer.__layer_name__)

	@verify_type('paranoid', envelope=WMessengerTextEnvelope, session=WMessengerOnionSessionProto)
	@verify_type(encoding=(str, None))
	def encode(self, envelope, session, encoding=None, **kwargs):
		""" :meth:`.WMessengerOnionCoderLayerProto.encode` method implementation.

		:param envelope: original envelope
		:param session: original session
		:param encoding: encoding to use (default is 'utf-8')
		:param kwargs: additional arguments

		:return: WMessengerBytesEnvelope
		"""
		message = envelope.message()
		message = message.encode() if encoding is None else message.encode(encoding)
		return WMessengerBytesEnvelope(message, meta=envelope)

	@verify_type('paranoid', envelope=WMessengerBytesEnvelope, session=WMessengerOnionSessionProto)
	@verify_type(encoding=(str, None))
	def decode(self, envelope, session, encoding=None, **kwargs):
		""" :meth:`.WMessengerOnionCoderLayerProto.decode` method implementation.

		:param envelope: original envelope
		:param session: original session
		:param encoding: encoding to use (default is 'utf-8')
		:param kwargs: additional arguments

		:return: WMessengerTextEnvelope
		"""
		message = envelope.message()
		message = message.decode() if encoding is None else message.decode(encoding)
		return WMessengerTextEnvelope(message, meta=envelope)


class WMessengerHexLayer(WMessengerOnionCoderLayerProto):
	""" :class:`.WMessengerOnionCoderLayerProto` class implementation. This class translate message to corresponding
	hex-string, or decodes it from hex-string to original binary representation.
	"""

	__layer_name__ = "com.binblob.wasp-general.hex-layer"
	""" Layer name
	"""

	def __init__(self):
		""" Construct new layer
		"""
		WMessengerOnionCoderLayerProto.__init__(self, WMessengerHexLayer.__layer_name__)

	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	@verify_type(envelope=WMessengerBytesEnvelope)
	def encode(self, envelope, session, **kwargs):
		""" :meth:`.WMessengerOnionCoderLayerProto.encode` method implementation.

		:param envelope: original envelope
		:param session: original session
		:param kwargs: additional arguments

		:return: WMessengerTextEnvelope
		"""
		return WMessengerTextEnvelope(str(WHex(envelope.message())), meta=envelope)

	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	@verify_type(envelope=WMessengerTextEnvelope)
	def decode(self, envelope, session, **kwargs):
		""" :meth:`.WMessengerOnionCoderLayerProto.decode` method implementation.

		:param envelope: original envelope
		:param session: original session
		:param kwargs: additional arguments

		:return: WMessengerBytesEnvelope
		"""
		return WMessengerBytesEnvelope(bytes(WUnHex(envelope.message())), meta=envelope)


class WMessengerBase64Layer(WMessengerOnionCoderLayerProto):
	""" :class:`.WMessengerOnionCoderLayerProto` class implementation. This class translate binary message
	to the corresponding base64 encoded bytes, or decodes it from base64 encoded bytes to the original binary
	representation.
	"""

	__layer_name__ = "com.binblob.wasp-general.base64-layer"
	""" Layer name
	"""

	def __init__(self):
		""" Construct new layer
		"""
		WMessengerOnionCoderLayerProto.__init__(self, WMessengerBase64Layer.__layer_name__)

	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	@verify_type(envelope=WMessengerBytesEnvelope)
	def encode(self, envelope, session, **kwargs):
		""" :meth:`.WMessengerOnionCoderLayerProto.encode` method implementation.

		:param envelope: original envelope
		:param session: original session
		:param kwargs: additional arguments

		:return: WMessengerBytesEnvelope
		"""
		return WMessengerBytesEnvelope(b64encode(envelope.message()), meta=envelope)

	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	@verify_type(envelope=WMessengerBytesEnvelope)
	def decode(self, envelope, session, **kwargs):
		""" :meth:`.WMessengerOnionCoderLayerProto.decode` method implementation.

		:param envelope: original envelope
		:param session: original session
		:param kwargs: additional arguments

		:return: WMessengerBytesEnvelope
		"""
		return WMessengerBytesEnvelope(b64decode(envelope.message()), meta=envelope)


class WMessengerAESLayer(WMessengerOnionCoderLayerProto):
	""" :class:`.WMessengerOnionCoderLayerProto` class implementation. This class encrypts/decrypts message with
	the specified AES cipher
	"""

	__layer_name__ = "com.binblob.wasp-general.aes-layer"
	""" Layer name
	"""

	def __init__(self):
		""" Construct new layer
		"""
		WMessengerOnionCoderLayerProto.__init__(self, WMessengerAESLayer.__layer_name__)

	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	@verify_type(envelope=WMessengerBytesEnvelope)
	@verify_type(aes_cipher=WAES)
	def encode(self, envelope, session, aes_cipher=None, **kwargs):
		""" :meth:`.WMessengerOnionCoderLayerProto.encode` method implementation.

		:param envelope: original envelope
		:param session: original session
		:param aes_cipher: cipher to use
		:param kwargs: additional arguments

		:return: WMessengerBytesEnvelope
		"""
		return WMessengerBytesEnvelope(aes_cipher.encrypt(envelope.message()), meta=envelope)

	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	@verify_type(envelope=WMessengerBytesEnvelope)
	@verify_type(aes_cipher=WAES)
	def decode(self, envelope, session, aes_cipher=None, **kwargs):
		""" :meth:`.WMessengerOnionCoderLayerProto.decode` method implementation.

		:param envelope: original envelope
		:param session: original session
		:param aes_cipher: cipher to use
		:param kwargs: additional arguments

		:return: WMessengerBytesEnvelope
		"""
		return WMessengerBytesEnvelope(aes_cipher.decrypt(envelope.message(), decode=False), meta=envelope)


class WMessengerRSALayer(WMessengerOnionCoderLayerProto):
	""" :class:`.WMessengerOnionCoderLayerProto` class implementation. This class encrypts/decrypts message with
	specified RSA cipher
	"""

	__layer_name__ = "com.binblob.wasp-general.rsa-layer"
	""" Layer name
	"""

	def __init__(self):
		""" Construct new layer
		"""
		WMessengerOnionCoderLayerProto.__init__(self, WMessengerRSALayer.__layer_name__)

	@verify_type('paranoid', session=WMessengerOnionSessionProto, public_key=WRSA.wrapped_class, sha_digest_size=int)
	@verify_type(envelope=WMessengerBytesEnvelope)
	def encode(self, envelope, session, public_key=None, sha_digest_size=32, **kwargs):
		""" :meth:`.WMessengerOnionCoderLayerProto.encode` method implementation.

		:param envelope: original envelope
		:param session: original session
		:param public_key: public key to encrypt
		:param sha_digest_size: SHA digest size to use
		:param kwargs: additional arguments

		:return: WMessengerBytesEnvelope
		"""
		message = WRSA.encrypt(envelope.message(), public_key, sha_digest_size=sha_digest_size)
		return WMessengerBytesEnvelope(message, meta=envelope)

	@verify_type('paranoid', session=WMessengerOnionSessionProto, private_key=WRSA.wrapped_class)
	@verify_type('paranoid', sha_digest_size=int)
	@verify_type(envelope=WMessengerBytesEnvelope)
	def decode(self, envelope, session, private_key=None, sha_digest_size=32, **kwargs):
		""" :meth:`.WMessengerOnionCoderLayerProto.decode` method implementation.

		:param envelope: original envelope
		:param session: original session
		:param private_key: private key to decrypt
		:param sha_digest_size: SHA digest size to use
		:param kwargs: additional arguments

		:return: WMessengerBytesEnvelope
		"""
		message = WRSA.decrypt(envelope.message(), private_key, sha_digest_size=sha_digest_size)
		return WMessengerBytesEnvelope(message, meta=envelope)



# -*- coding: utf-8 -*-

import pytest
from datetime import datetime, timedelta
from decorator import decorator
from threading import Event

from wasp_general.datetime import utc_datetime
from wasp_general.task.thread import WThreadTask
from wasp_general.task.scheduler.proto import WScheduleTask, WScheduleRecord
from wasp_general.task.scheduler.proto import WRunningRecordRegistryProto, WTaskSourceProto, WSchedulerServiceProto

from wasp_general.task.scheduler.scheduler import WSchedulerWatchdog, WRunningRecordRegistry, WPostponedRecordRegistry
from wasp_general.task.scheduler.scheduler import WTaskSourceRegistry, WSchedulerService


def repeat_fn(count):
	def fl_decorator(decorated_fn):
		def sl_decorator(original_fn, *args, **kwargs):
			for i in range(count):
				original_fn(*args, **kwargs)
		return decorator(sl_decorator)(decorated_fn)
	return fl_decorator


__test_repeat_count__ = 25


class TestWSchedulerWatchdog:

	class HFWatchdog(WSchedulerWatchdog):
		__thread_polling_timeout__ = 0.01

	class DummyTask(WScheduleTask):

		__thread_polling_timeout__ = 0.01

		def __init__(self, wait_for=None):
			WScheduleTask.__init__(self)
			self.started = Event()
			self.wait_for = wait_for

		def thread_started(self):
			self.started.set()
			if self.wait_for is None:
				return

			while self.wait_for.is_set() is False and self.stop_event().is_set() is False:
				self.wait_for.wait(TestWSchedulerWatchdog.DummyTask.__thread_polling_timeout__)

		def thread_stopped(self):
			self.started.clear()

	@repeat_fn(__test_repeat_count__)
	def test(self):
		task = TestWSchedulerWatchdog.DummyTask()
		schedule = WScheduleRecord(task)

		registry = WRunningRecordRegistry()

		pytest.raises(TypeError, WSchedulerWatchdog.create, schedule, 1)

		dog = WSchedulerWatchdog.create(schedule, registry)
		assert(isinstance(dog, WSchedulerWatchdog) is True)
		assert(isinstance(dog, WThreadTask) is True)
		assert(dog.record() == schedule)
		assert(dog.registry() == registry)

		stop_event = Event()
		task = TestWSchedulerWatchdog.DummyTask(stop_event)
		schedule = WScheduleRecord(task)
		dog = TestWSchedulerWatchdog.HFWatchdog.create(schedule, registry)
		dog.start()
		dog.start_event().wait()
		task.started.wait()
		assert(dog.record() == schedule)
		pytest.raises(RuntimeError, dog.start)
		stop_event.set()
		dog.stop()

		buggy_schedule = WScheduleRecord(task)
		dog = WSchedulerWatchdog.create(buggy_schedule, registry)
		pytest.raises(RuntimeError, dog.thread_started)

		buggy_schedule = WScheduleRecord(task)

		class A:
			def uid(self):
				return 1

		buggy_schedule.task = lambda: A()
		dog = WSchedulerWatchdog.create(buggy_schedule, registry)
		pytest.raises(RuntimeError, dog.start)


class TestWRunningRecordRegistry:

	class HFRunningTaskRegistry(WRunningRecordRegistry):
		__thread_polling_timeout__ = TestWSchedulerWatchdog.HFWatchdog.__thread_polling_timeout__ / 2

	@repeat_fn(__test_repeat_count__)
	def test(self):
		registry = WRunningRecordRegistry(thread_name_suffix='!')
		assert(isinstance(registry, WRunningRecordRegistry) is True)
		assert(isinstance(registry, WRunningRecordRegistryProto) is True)
		assert (isinstance(registry, WThreadTask) is True)
		assert(registry.watchdog_class() == WSchedulerWatchdog)

		registry = WRunningRecordRegistry(watchdog_cls=TestWSchedulerWatchdog.HFWatchdog)
		assert(registry.watchdog_class() == TestWSchedulerWatchdog.HFWatchdog)

		registry = TestWRunningRecordRegistry.HFRunningTaskRegistry()
		registry.start()
		registry.start_event().wait()
		assert(len(registry) == 0)
		assert(registry.running_records() == tuple())

		task1_stop_event = Event()
		task = TestWSchedulerWatchdog.DummyTask(task1_stop_event)
		schedule = WScheduleRecord(task)
		registry.exec(schedule)
		task.started.wait()
		assert(len(registry) == 1)
		running_task = registry.running_records()
		assert(isinstance(running_task, tuple) is True)
		assert(len(running_task) == 1)

		running_task = running_task[0]
		assert(running_task == schedule)

		task1_stop_event.set()
		task.ready_event().wait()

		task1_stop_event = Event()
		task = TestWSchedulerWatchdog.DummyTask(task1_stop_event)
		schedule = WScheduleRecord(task)
		registry.exec(schedule)
		task.start_event().wait()
		task.started.wait()
		assert(len(registry) >= 1)
		registry.stop()


class TestWPostponedRecordRegistry:

	@repeat_fn(__test_repeat_count__)
	def test(self):
		registry = WPostponedRecordRegistry()
		assert(registry.maximum_records() is None)
		assert(registry.has_records() is False)
		assert(len(registry) == 0)

		drop_count = []

		def on_drop():
			drop_count.append(None)

		task = TestWSchedulerWatchdog.DummyTask()
		wait_schedule1 = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.wait, on_drop=on_drop
		)
		wait_schedule2 = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.wait, on_drop=on_drop
		)

		assert(len(drop_count) == 0)
		assert(len(registry) == 0)

		registry.postpone(wait_schedule1)
		assert(len(drop_count) == 0)
		assert(len(registry) == 1)

		registry.postpone(wait_schedule2)
		assert(len(drop_count) == 0)
		assert(len(registry) == 2)

		tasks = [x for x in registry]
		assert(len(registry) == 0)
		assert(wait_schedule1 in tasks)
		assert(wait_schedule2 in tasks)

		registry = WPostponedRecordRegistry(maximum_records=0)
		assert(len(drop_count) == 0)
		assert(len(registry) == 0)

		registry.postpone(wait_schedule1)
		assert(len(drop_count) == 1)
		assert(len(registry) == 0)

		registry.postpone(wait_schedule2)
		assert(len(drop_count) == 2)
		assert(len(registry) == 0)

		drop_count.clear()
		registry = WPostponedRecordRegistry(maximum_records=1)
		assert(len(drop_count) == 0)
		assert(len(registry) == 0)

		registry.postpone(wait_schedule1)
		assert(len(drop_count) == 0)
		assert(len(registry) == 1)

		registry.postpone(wait_schedule2)
		assert(len(drop_count) == 1)
		assert(len(registry) == 1)

		tasks = [x for x in registry]
		assert(len(registry) == 0)
		assert(wait_schedule1 in tasks)
		assert(wait_schedule2 not in tasks)

		drop_count.clear()
		registry = WPostponedRecordRegistry()
		assert(len(drop_count) == 0)
		assert(len(registry) == 0)

		drop_schedule = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.drop, on_drop=on_drop
		)
		registry.postpone(drop_schedule)
		assert(len(drop_count) == 1)
		assert(len(registry) == 0)

		drop_count.clear()
		postpone_first_group_1_schedule1 = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.postpone_first, task_group_id='group1', on_drop=on_drop
		)

		postpone_first_group_1_schedule2 = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.postpone_first, task_group_id='group1', on_drop=on_drop
		)

		postpone_first_group_2_schedule = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.postpone_first, task_group_id='group2', on_drop=on_drop
		)

		postpone_first_schedule1 = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.postpone_first, on_drop=on_drop
		)

		postpone_first_schedule2 = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.postpone_first, on_drop=on_drop
		)

		registry.postpone(postpone_first_group_1_schedule1)
		assert(len(drop_count) == 0)
		assert(len(registry) == 1)

		registry.postpone(postpone_first_group_1_schedule2)
		assert(len(drop_count) == 1)
		assert(len(registry) == 1)

		registry.postpone(postpone_first_group_2_schedule)
		assert(len(drop_count) == 1)
		assert(len(registry) == 2)

		registry.postpone(postpone_first_schedule1)
		assert(len(drop_count) == 1)
		assert(len(registry) == 3)

		registry.postpone(postpone_first_schedule2)
		assert(len(drop_count) == 1)
		assert(len(registry) == 4)

		tasks = [x for x in registry]
		assert(len(registry) == 0)
		assert(postpone_first_group_1_schedule1 in tasks)
		assert(postpone_first_group_2_schedule in tasks)
		assert(postpone_first_schedule1 in tasks)
		assert(postpone_first_schedule2 in tasks)

		wait_group_1 = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.wait, task_group_id='group1', on_drop=on_drop
		)
		registry.postpone(wait_group_1)
		pytest.raises(RuntimeError, registry.postpone, postpone_first_group_1_schedule1)
		tasks = [x for x in registry]

		drop_count.clear()
		postpone_last_group_1_schedule1 = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.postpone_last, task_group_id='group1', on_drop=on_drop
		)

		postpone_last_group_1_schedule2 = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.postpone_last, task_group_id='group1', on_drop=on_drop
		)

		postpone_last_group_2_schedule = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.postpone_last, task_group_id='group2', on_drop=on_drop
		)

		postpone_last_schedule1 = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.postpone_last, on_drop=on_drop
		)

		postpone_last_schedule2 = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.postpone_last, on_drop=on_drop
		)
		registry.postpone(postpone_last_group_1_schedule1)
		assert(len(drop_count) == 0)
		assert(len(registry) == 1)

		registry.postpone(postpone_last_group_1_schedule2)
		assert(len(drop_count) == 1)
		assert(len(registry) == 1)

		registry.postpone(postpone_last_group_2_schedule)
		assert(len(drop_count) == 1)
		assert(len(registry) == 2)

		registry.postpone(postpone_last_schedule1)
		assert(len(drop_count) == 1)
		assert(len(registry) == 3)

		registry.postpone(postpone_last_schedule2)
		assert(len(drop_count) == 1)
		assert(len(registry) == 4)

		tasks = [x for x in registry]
		assert(len(registry) == 0)
		assert(postpone_last_group_1_schedule2 in tasks)
		assert(postpone_last_group_2_schedule in tasks)
		assert(postpone_last_schedule1 in tasks)
		assert(postpone_last_schedule2 in tasks)

		wait_group_1 = WScheduleRecord(
			task, policy=WScheduleRecord.PostponePolicy.wait, task_group_id='group1', on_drop=on_drop
		)
		registry.postpone(wait_group_1)
		pytest.raises(RuntimeError, registry.postpone, postpone_last_group_1_schedule1)

		wait_group_1.policy = lambda: None
		pytest.raises(RuntimeError, registry.postpone, wait_group_1)


class TestWTaskSourceRegistry:

	class TaskSource(WTaskSourceProto):

		def __init__(self):
			WTaskSourceProto.__init__(self)
			self.tasks = []

		def has_records(self):
			if self.tasks is not None:
				result = tuple(self.tasks)
				self.tasks.clear()
				return result

		def next_start(self):
			if len(self.tasks) > 0:
				return utc_datetime()

		def tasks_planned(self):
			return len(self.tasks)

		def scheduler_service(self):
			return None

	@repeat_fn(__test_repeat_count__)
	def test(self):
		registry = WTaskSourceRegistry()
		assert(registry.check() is None)
		assert(registry.task_sources() == tuple())

		task_source1 = TestWTaskSourceRegistry.TaskSource()
		registry.add_source(task_source1)
		assert(registry.check() is None)
		assert(registry.task_sources() == (task_source1, ))

		task1 = WScheduleRecord(TestWSchedulerWatchdog.DummyTask())
		task_source1.tasks.append(task1)
		registry.update()
		assert(registry.check() == (task1, ))
		assert(registry.check() is None)

		task_source2 = TestWTaskSourceRegistry.TaskSource()
		registry.add_source(task_source2)
		assert(registry.check() is None)
		result = registry.task_sources()
		assert(result == (task_source1, task_source2) or result == (task_source2, task_source1))

		task_source1.tasks.append(task1)
		task2 = WScheduleRecord(TestWSchedulerWatchdog.DummyTask())
		task_source2.tasks.append(task2)
		assert(registry.check() is None)

		registry.update(task_source=task_source2)
		assert(registry.check() == (task2, ))
		assert(registry.check() == (task1, ))

		utc_now = utc_datetime()
		task_source1.tasks.append(task1)
		task_source2.tasks.append(task2)
		task_source2.next_start = lambda: utc_now
		registry.update()
		assert(registry.check() == (task2, ))

		task_source1.next_start = lambda: utc_now
		task_source2.tasks.append(task2)
		registry.update()
		result = registry.check()
		assert(result == (task1, task2) or result == (task2, task1))

		task_source1.next_start = lambda: datetime.now()
		pytest.raises(ValueError, registry.update)


class TestWSchedulerService:

	__wait_task_timeout__ = 0.001

	class HFSchedulerService(WSchedulerService):
		__thread_polling_timeout__ = (
			TestWRunningRecordRegistry.HFRunningTaskRegistry.__thread_polling_timeout__ / 2
		)

	class DummyTask(TestWSchedulerWatchdog.DummyTask):

		__result__ = 0
		__dropped__ = 0
		__waited__ = 0

		def __init__(self, wait_for=None):
			TestWSchedulerWatchdog.DummyTask.__init__(self, wait_for=wait_for)
			self.drop_event = Event()
			self.wait_event = Event()

		def thread_started(self):
			TestWSchedulerWatchdog.DummyTask.thread_started(self)
			TestWSchedulerService.DummyTask.__result__ += 1

		def on_drop(self):
			TestWSchedulerService.DummyTask.__dropped__ += 1
			self.drop_event.set()

		def on_wait(self):
			TestWSchedulerService.DummyTask.__waited__ += 1
			self.wait_event.set()

	def wait_for_events(*events, every=False):
		events = list(events)
		while len(events) > 0:

			for i in range(len(events)):
				event = events[i]
				if event.is_set() is True:
					if every is False:
						return
					events.pop(i)
					break

			if len(events) > 0:
				events[0].wait(TestWSchedulerService.__wait_task_timeout__)

	@staticmethod
	def wait_for_tasks(*tasks, every=False):
		tasks = list(tasks)
		while len(tasks) > 0:

			for i in range(len(tasks)):
				task = tasks[i]
				if task.check_events() is True or task.drop_event.is_set() is True:
					if every is False:
						return
					tasks.pop(i)
					break

			if len(tasks) > 0:
				tasks[0].ready_event().wait(TestWSchedulerService.__wait_task_timeout__)

	@repeat_fn(__test_repeat_count__)
	def test(self):
		TestWSchedulerService.DummyTask.__result__ = 0
		TestWSchedulerService.DummyTask.__dropped__ = 0
		TestWSchedulerService.DummyTask.__waited__ = 0

		service = WSchedulerService(thread_name_suffix='!')
		assert(isinstance(service, WSchedulerService) is True)
		assert(isinstance(service, WSchedulerServiceProto) is True)
		assert(isinstance(service, WThreadTask) is True)
		assert(service.maximum_running_records() > 0)
		assert(service.maximum_running_records() == WSchedulerService.__default_maximum_running_records__)
		assert(service.maximum_postponed_records() is None)
		assert(service.task_sources() == tuple())

		service = WSchedulerService(maximum_postponed_records=2, maximum_running_records=1)
		assert(service.maximum_running_records() == 1)
		assert(service.maximum_postponed_records() == 2)

		pytest.raises(
			ValueError, WSchedulerService, maximum_postponed_records=1,
			postponed_record_registry=WPostponedRecordRegistry()
		)

		service = TestWSchedulerService.HFSchedulerService(
			maximum_running_records=2,
			running_record_registry=TestWRunningRecordRegistry.HFRunningTaskRegistry(
				watchdog_cls=TestWSchedulerWatchdog.HFWatchdog
			)
		)

		task_source1 = TestWTaskSourceRegistry.TaskSource()
		service.add_task_source(task_source1)
		assert(service.task_sources() == (task_source1, ))

		service.start()
		service.start_event().wait()

		assert (TestWSchedulerService.DummyTask.__result__ == 0)
		assert(TestWSchedulerService.DummyTask.__dropped__ == 0)
		assert(TestWSchedulerService.DummyTask.__waited__ == 0)
		task1 = TestWSchedulerService.DummyTask()
		task_source1.tasks.append(WScheduleRecord(task1, on_drop=task1.on_drop, on_wait=task1.on_wait))
		service.update()

		task1.start_event().wait()

		TestWSchedulerService.wait_for_tasks(task1)

		assert(TestWSchedulerService.DummyTask.__result__ == 1)
		assert(TestWSchedulerService.DummyTask.__dropped__ == 0)
		assert(TestWSchedulerService.DummyTask.__waited__ == 0)

		task_source2 = TestWTaskSourceRegistry.TaskSource()
		service.add_task_source(task_source2)
		result = service.task_sources()
		assert(result == (task_source1, task_source2) or result == (task_source2, task_source1))

		long_run_task = TestWSchedulerService.DummyTask(Event())

		group1_task1_stop_event = Event()
		group1_task1 = TestWSchedulerService.DummyTask(group1_task1_stop_event)
		group1_task2_stop_event = Event()
		group1_task2 = TestWSchedulerService.DummyTask(group1_task2_stop_event)
		task_source1.tasks.append(WScheduleRecord(
			long_run_task, on_drop=long_run_task.on_drop, on_wait=long_run_task.on_wait
		))
		task_source1.tasks.append(
			WScheduleRecord(
				group1_task1, on_drop=group1_task1.on_drop, on_wait=group1_task1.on_wait,
				task_group_id='group1', policy=WScheduleRecord.PostponePolicy.drop
			)
		)
		task_source2.tasks.append(
			WScheduleRecord(
				group1_task2, on_drop=group1_task2.on_drop, on_wait=group1_task2.on_wait,
				task_group_id='group1', policy=WScheduleRecord.PostponePolicy.drop
			)
		)

		service.update()

		TestWSchedulerService.wait_for_events(group1_task1.start_event(), group1_task2.start_event())

		result = service.records_status()
		assert(result == (1, 0) or result == (2, 0))

		running_tasks = service.running_records()
		for record in running_tasks:
			task = record.task()
			assert(task in (group1_task1, group1_task2, long_run_task))

		TestWSchedulerService.wait_for_tasks(group1_task1, group1_task2)

		group1_task1_stop_event.set()
		group1_task2_stop_event.set()
		TestWSchedulerService.wait_for_tasks(group1_task1, group1_task2, every=True)
		assert(TestWSchedulerService.DummyTask.__result__ == 2)
		assert(TestWSchedulerService.DummyTask.__dropped__ == 1)
		assert(TestWSchedulerService.DummyTask.__waited__ == 0)

		group1_task1.stop()
		group1_task2.stop()

		group1_task1_stop_event = Event()
		group1_task2_stop_event = Event()

		group1_task1 = TestWSchedulerService.DummyTask(group1_task1_stop_event)
		group1_task2 = TestWSchedulerService.DummyTask(group1_task2_stop_event)

		task_source1.tasks.append(
			WScheduleRecord(
				group1_task1, on_drop=group1_task1.on_drop, on_wait=group1_task1.on_wait,
				task_group_id='group1', policy=WScheduleRecord.PostponePolicy.wait
			)
		)
		task_source2.tasks.append(
			WScheduleRecord(
				group1_task2, on_drop=group1_task2.on_drop, on_wait=group1_task2.on_wait,
				task_group_id='group1', policy=WScheduleRecord.PostponePolicy.wait
			)
		)

		service.update()

		result = service.records_status()

		group1_task1_stop_event.set()
		group1_task2_stop_event.set()
		TestWSchedulerService.wait_for_tasks(group1_task1, group1_task2, every=True)

		group1_task1.wait_event.wait()
		group1_task2.wait_event.wait()

		assert(TestWSchedulerService.DummyTask.__result__ == 4)
		assert(TestWSchedulerService.DummyTask.__dropped__ == 1)
		assert(TestWSchedulerService.DummyTask.__waited__ == 2)

		service.stop()
		TestWSchedulerService.wait_for_tasks(long_run_task)



# -*- coding: utf-8 -*-
# wasp_general/uri.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: merge some from wasp_general.network.web.service and wasp_general.network.web.re_statements

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from enum import Enum
from urllib.parse import urlsplit, urlunsplit
from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type, verify_subclass


class WURI:
	"""
	Class that represent URI as it is described in RFC 3986
	"""

	class Component(Enum):
		""" Parts/components names that URI is consists of
		"""
		scheme = 'scheme'
		username = 'username'
		password = 'password'
		hostname = 'hostname'
		port = 'port'
		path = 'path'
		query = 'query'
		fragment = 'fragment'

	def __init__(self, **components):
		""" Create new WURI object. By default empty URI is created

		:param components: components that must be set for this URI. Keys - components names as they
		defined in :class:`.WURI.Component`, values - corresponding values
		"""
		self.__components = {x: None for x in WURI.Component}

		for component_name, component_value in components.items():
			self.component(component_name, component_value)

	@verify_type(item=str)
	def __getattr__(self, item):
		""" Return component value by its name

		:param item: component name
		:return: str
		"""
		try:
			components_fn = object.__getattribute__(self, WURI.component.__name__)
			item = WURI.Component(item)
			return lambda: components_fn(item)
		except ValueError:
			pass

		return object.__getattribute__(self, item)

	def __str__(self):
		""" Return string that represents this URI

		:return: str
		"""
		netloc = ''

		username = self.username()
		if username is not None:
			netloc += username

		password = self.password()
		if password is not None:
			netloc += ':' + password

		if len(netloc) > 0:
			netloc += '@'

		hostname = self.hostname()
		if hostname is not None:
			netloc += hostname

		port = self.port()
		if port is not None:
			netloc += ':' + str(port)

		scheme = self.scheme()
		path = self.path()
		if len(netloc) == 0 and scheme is not None and path is not None:
			path = '//' + path

		return urlunsplit((
			scheme if scheme is not None else '',
			netloc,
			path if path is not None else '',
			self.query(),
			self.fragment()
		))

	@verify_type(component=(str, Component))
	def component(self, component, value=None):
		""" Set and/or get component value.

		:param component: component name to return
		:param value: if value is not None, this value will be set as a component value
		:return: str
		"""
		if isinstance(component, str) is True:
			component = WURI.Component(component)
		if value is not None:
			self.__components[component] = value
			return value
		return self.__components[component]

	@classmethod
	@verify_type(uri=str)
	def parse(cls, uri):
		""" Parse URI-string and return WURI object

		:param uri: string to parse
		:return: WURI
		"""
		uri_components = urlsplit(uri)
		adapter_fn = lambda x: x if x is not None and (isinstance(x, str) is False or len(x)) > 0 else None

		return cls(
			scheme=adapter_fn(uri_components.scheme),
			username=adapter_fn(uri_components.username),
			password=adapter_fn(uri_components.password),
			hostname=adapter_fn(uri_components.hostname),
			port=adapter_fn(uri_components.port),
			path=adapter_fn(uri_components.path),
			query=adapter_fn(uri_components.query),
			fragment=adapter_fn(uri_components.fragment),
		)

	def __iter__(self):
		""" Iterate over URI components. This method yields tuple of component name and its value

		:return: generator
		"""
		for component in WURI.Component:
			component_name = component.value
			component_value_fn = getattr(self, component_name)
			yield component, component_value_fn()


class WSchemeSpecification:
	""" Specification for URI, that is described by scheme-component
	"""

	class ComponentDescriptor(Enum):
		""" Value that describes component relation to a scheme specification
		"""
		required = 0
		optional = 1
		unsupported = None

	@verify_type(scheme_name=str)
	def __init__(self, scheme_name, **descriptors):
		""" Create new scheme specification. Every component that was not described by this method is treated
		as unsupported

		:param scheme_name: URI scheme value
		:param descriptors: component names and its descriptors
		(:class:`.WSchemeSpecification.ComponentDescriptor`)
		"""
		self.__scheme_name = scheme_name

		self.__descriptors = {x: WSchemeSpecification.ComponentDescriptor.unsupported for x in WURI.Component}
		self.__descriptors[WURI.Component.scheme] = WSchemeSpecification.ComponentDescriptor.required

		for descriptor_name in descriptors.keys():
			component = WURI.Component(descriptor_name)
			if component == WURI.Component.scheme:
				raise TypeError('Scheme name can not be specified twice')
			descriptor = descriptors[descriptor_name]
			if isinstance(descriptor, WSchemeSpecification.ComponentDescriptor) is False:
				raise TypeError('Invalid "%s" descriptor type' % descriptor_name)
			self.__descriptors[component] = descriptor

	def scheme_name(self):
		""" Return scheme name that this specification is describing

		:return: str
		"""
		return self.__scheme_name

	@verify_type(component=WURI.Component)
	def descriptor(self, component):
		""" Return descriptor for the specified component

		:param component: component name which descriptor should be returned
		:return: WSchemeSpecification.ComponentDescriptor
		"""
		return self.__descriptors[component]

	def __iter__(self):
		""" Iterate over URI components. This method yields tuple of component (:class:`.WURI.Component`) and
		its descriptor

		:return: generator
		"""
		for component in WURI.Component:
			yield component, self.__descriptors[component]

	@verify_type(uri=WURI)
	def is_compatible(self, uri):
		""" Check if URI is compatible with this specification. Compatible URI has scheme name that matches
		specification scheme name, has all of the required components, does not have unsupported components
		and may have optional components

		:param uri: URI to check
		:return: bool
		"""
		for component, component_value in uri:
			descriptor = self.descriptor(component)
			if component_value is None:
				if descriptor == WSchemeSpecification.ComponentDescriptor.required:
					return False
			elif descriptor == WSchemeSpecification.ComponentDescriptor.unsupported:
					return False
		return True


class WSchemeHandler(metaclass=ABCMeta):
	""" Handler that do some work for compatible URI
	"""

	@classmethod
	@abstractmethod
	def scheme_specification(cls):
		""" Return scheme specification

		:return: WSchemeSpecification
		"""
		raise NotImplementedError('This method is abstract')

	@classmethod
	@abstractmethod
	@verify_type(uri=WURI)
	def create_handler(cls, uri, **kwargs):
		""" Return handler instance

		:param uri: original URI, that a handler is created for
		:param kwargs: additional arguments that may be used by a handler specialization

		:return: WSchemeHandler
		"""
		raise NotImplementedError('This method is abstract')


class WSchemeCollection:
	""" Collection of URI scheme handlers, that is capable to process different WURI. Only one handler per scheme
	is supported. Suitable handler will be searched by a scheme name.
	"""

	class NoHandlerFound(Exception):
		""" Exception that is raised when no handler is found for a URI
		"""

		def __init__(self, uri):
			""" Create new exception object

			:param uri: URI which scheme does not have a corresponding handler
			"""
			Exception.__init__(self, 'No handler was found for the specified URI: %s' % str(uri))

	class SchemeIncompatible(Exception):
		""" Exception that is raised when URI does not match a specification (:class:`.WSchemeSpecification`).
		This happens if URI has unsupported components or does not have a required components.
		"""

		def __init__(self, uri):
			""" Create new exception object

			:param uri: URI that does not comply a handler specification (:class:`.WSchemeSpecification`)
			"""
			Exception.__init__(
				self,
				'Handler was found for the specified scheme. '
				'But URI has not required components or has unsupported components: %s' % str(uri)
			)

	def __init__(self, *scheme_handlers_cls, default_handler_cls=None):
		""" Create new collection

		:param scheme_handlers_cls: handlers to add to this collection
		:param default_handler: handler that must be called for a URI that does not have scheme component
		"""
		self.__handlers_cls = []
		self.__default_handler_cls = default_handler_cls
		for handler_cls in scheme_handlers_cls:
			self.add(handler_cls)

	@verify_subclass(scheme_handler_cls=WSchemeHandler)
	def add(self, scheme_handler_cls):
		""" Append the specified handler to this collection

		:param scheme_handler_cls: handler that should be added
		:return: None
		"""
		self.__handlers_cls.append(scheme_handler_cls)

	def handler(self, scheme_name=None):
		""" Return handler which scheme name matches the specified one

		:param scheme_name: scheme name to search for
		:return: WSchemeHandler class or None (if matching handler was not found)
		"""
		if scheme_name is None:
			return self.__default_handler_cls
		for handler in self.__handlers_cls:
			if handler.scheme_specification().scheme_name() == scheme_name:
				return handler

	@verify_type(uri=WURI)
	def open(self, uri, **kwargs):
		""" Return handler instance that matches the specified URI. WSchemeCollection.NoHandlerFound and
		WSchemeCollection.SchemeIncompatible may be raised.

		:param uri: URI to search handler for
		:param kwargs: additional arguments that may be used by a handler specialization
		:return: WSchemeHandler
		"""
		handler = self.handler(uri.scheme())
		if handler is None:
			raise WSchemeCollection.NoHandlerFound(uri)

		if uri.scheme() is None:
			uri.component('scheme', handler.scheme_specification().scheme_name())

		if handler.scheme_specification().is_compatible(uri) is False:
			raise WSchemeCollection.SchemeIncompatible(uri)

		return handler.create_handler(uri, **kwargs)



# -*- coding: utf-8 -*-
# wasp_general/network/messenger/envelope.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.


# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from wasp_general.verify import verify_type

from wasp_general.network.messenger.proto import WMessengerEnvelopeProto


class WMessengerEnvelope(WMessengerEnvelopeProto):
	""" Simple :class:`.WMessengerEnvelopeProto` implementation. "Meta-dictionary" is used as is without any
	restrictions or limitations, so any further envelopes (envelopes that uses this envelope as meta) can rewrite
	or remove saved information
	"""

	@verify_type(meta=(WMessengerEnvelopeProto, dict, None))
	def __init__(self, data, meta=None):
		""" Construct new envelope

		:param data: original message
		:param meta: envelope meta-data to copy
		"""
		self.__data = data
		self.__meta = {}
		if meta is not None:
			if isinstance(meta, WMessengerEnvelopeProto) is True:
				if isinstance(meta, WMessengerEnvelope) is False:
					raise TypeError('meta must be WMessengerEnvelope (or derived classes) object')
				self.__meta = meta.meta()
			elif isinstance(meta, dict):
				self.__meta = meta.copy()

	def message(self):
		""" Return original message

		:return: any type
		"""
		return self.__data

	def meta(self):
		""" Return envelope dictionary copy

		:return: dict
		"""
		return self.__meta.copy()

	@verify_type(key=str)
	def add_meta(self, key, value):
		""" Add meta-information (value) for the given key

		:param key: meta-key
		:param value: meta-value
		:return: None
		"""
		self.__meta[key] = value


class WMessengerTextEnvelope(WMessengerEnvelope):
	""" Envelope for str-objects
	"""

	@verify_type(data=str, meta=(WMessengerEnvelopeProto, dict, None))
	def __init__(self, data, meta=None):
		WMessengerEnvelope.__init__(self, data, meta=meta)


class WMessengerBytesEnvelope(WMessengerEnvelope):
	""" Envelope for bytes-objects
	"""

	@verify_type(data=bytes, meta=(WMessengerEnvelopeProto, dict, None))
	def __init__(self, data, meta=None):
		WMessengerEnvelope.__init__(self, data, meta=meta)


class WMessengerDictEnvelope(WMessengerEnvelope):
	""" Envelope for dict-objects (dictionary objects)
	"""

	@verify_type(data=dict, meta=(WMessengerEnvelopeProto, dict, None))
	def __init__(self, data, meta=None):
		WMessengerEnvelope.__init__(self, data, meta=meta)



# -*- coding: utf-8 -*-
# wasp_general/os/linux/lvm.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import subprocess
import os
import math

from wasp_general.verify import verify_type, verify_value
from wasp_general.os.linux.mounts import WMountPoint


class WLVMInfoCommand:
	""" This is a helper, with which it is easier to call for pvdisplay, vgdisplay or lvdisplay program.

	This class uses subprocess.check_output method for a program calling. And when non-zero code is returned by
	the program, an subprocess.CalledProcessError exception is raised. There is a timeout for a program to be
	complete. If a program wasn't completed for that period of time, subprocess.TimeoutExpired exception is
	raised
	"""

	__lvm_cmd_default_timeout__ = 3
	""" Default timeout for command to process
	"""

	@verify_type(command=str, fields_count=int, cmd_timeout=(int, float, None), sudo=bool)
	@verify_value(cmd_timeout=lambda x: x is None or x > 0)
	def __init__(self, command, fields_count, cmd_timeout=None, sudo=False):
		""" Create new command

		:param command: program to execute
		:param fields_count: fields in a program output
		:param cmd_timeout: timeout for a program (if it is None - then default value is used)
		:param sudo: flag - whether to run this program with sudo or not
		"""
		self.__command = command
		self.__fields_count = fields_count
		self.__cmd_timeout = cmd_timeout if cmd_timeout is not None else self.__lvm_cmd_default_timeout__
		self.__sudo = sudo

	def command(self):
		""" Return target program

		:return: str
		"""
		return self.__command

	def fields_count(self):
		""" Return number of fields in a program output

		:return: int
		"""
		return self.__fields_count

	def cmd_timeout(self):
		""" Timeout for a program to complete

		:return: int, float
		"""
		return self.__cmd_timeout

	def sudo(self):
		""" Return 'sudo' flag (whether to run this program with sudo or not)

		:return: bool
		"""
		return self.__sudo

	@verify_type(name=(str, None))
	def lvm_info(self, name=None):
		""" Call a program

		:param name: if specified - program will return information for that lvm-entity only. otherwise -
		all available entries are returned

		:return: tuple of str (fields)
		"""
		cmd = [] if self.sudo() is False else ['sudo']
		cmd.extend([self.command(), '-c'])
		if name is not None:
			cmd.append(name)

		output = subprocess.check_output(cmd, timeout=self.cmd_timeout())
		output = output.decode()
		result = []
		fields_count = self.fields_count()
		for line in output.split('\n'):
			line = line.strip()
			fields = line.split(':')
			if len(fields) == fields_count:
				result.append(fields)
		if name is not None and len(result) != 1:
			raise RuntimeError('Unable to parse command result')
		return tuple(result)


class WLVMInfo:
	""" Basic class for actual LVM information. This class creates :class:`.WLVMInfoCommand` object
	which may be called on an object creation (it depends on constructor parameters)
	"""

	__lvm_info_cmd_timeout__ = 3
	""" Timeout for a program to complete
	"""

	@verify_type('paranoid', command=str, fields_count=int, sudo=bool)
	@verify_type(lvm_entity=(str, tuple, list, set))
	@verify_value(lvm_entity=lambda x: len(x) > 0 if isinstance(x, str) else True)
	def __init__(self, command, fields_count, lvm_entity, sudo=False):
		""" Create new info-object

		:param command: same as command in :meth:`.WLVMInfoCommand.__init__`
		:param fields_count: same as fields_count in :meth:`.WLVMInfoCommand.__init__`
		:param lvm_entity: if this is a list/tuple/set - then it is a collection of fields (collection length \
		must be the same as 'fields_count'). If it is a string, then command is executed to get corresponding \
		fields
		:param sudo: same as sudo in :meth:`.WLVMInfoCommand.__init__`
		"""
		self.__lvm_command = WLVMInfoCommand(
			command, fields_count, cmd_timeout=self.__class__.__lvm_info_cmd_timeout__, sudo=sudo
		)
		if isinstance(lvm_entity, (tuple, list, set)) is True:
			if len(lvm_entity) != fields_count:
				raise ValueError(
					'Invalid lvm entity fields count: %i (expected: %i)' %
					(len(lvm_entity), fields_count)
				)
			self.__lvm_entity = tuple(lvm_entity)
		else:
			self.__lvm_entity = (self.lvm_command().lvm_info(lvm_entity)[0])

	def lvm_command(self):
		""" Return LVM-command object

		:return: WLVMInfoCommand
		"""
		return self.__lvm_command

	def lvm_entity(self):
		""" Return object fields

		:return: tuple of str (fields)
		"""
		return self.__lvm_entity


class WPhysicalVolume(WLVMInfo):
	""" Class represent a physical volume
	"""

	@verify_type('paranoid', physical_volume=(str, tuple, list, set), sudo=bool)
	@verify_value('paranoid', physical_volume=lambda x: len(x) > 0 if isinstance(x, str) else True)
	def __init__(self, physical_volume, sudo=False):
		""" Create new physical volume descriptor

		:param physical_volume: same as 'lvm_entity' in :meth:`.WLVMInfo.__init__`
		:param sudo: same as 'sudo' in :meth:`.WLVMInfo.__init__`
		"""
		WLVMInfo.__init__(self, 'pvdisplay', 12, physical_volume, sudo=sudo)

	def all(self):
		""" Return every physical volume in the system

		:return: tuple of WPhysicalVolume
		"""
		return tuple([WPhysicalVolume(x) for x in self.lvm_command().lvm_info()])

	def device_name(self):
		""" Return physical volume device name

		:return: str
		"""
		return self.lvm_entity()[0]

	def volume_group(self):
		""" Return related volume group name (may be empty string if this volume is not allocated to any)

		:return: str
		"""
		return self.lvm_entity()[1]

	def sectors_count(self):
		""" Return physical volume size in sectors

		:return: int
		"""
		return int(self.lvm_entity()[2])

	def extent_size(self):
		""" Return physical extent size in kilobytes (may have 0 value if this volume is not allocated to any)

		:return: int
		"""
		return int(self.lvm_entity()[7])

	def total_extents(self):
		""" Return total number of physical extents (may have 0 value if this volume is not allocated to any)

		:return: int
		"""
		return int(self.lvm_entity()[8])

	def free_extents(self):
		""" Return free number of physical extents (may have 0 value if this volume is not allocated to any)

		:return: int
		"""
		return int(self.lvm_entity()[9])

	def allocated_extents(self):
		""" Return allocated number of physical extents (may have 0 value if this volume is not allocated to \
		any)

		:return: int
		"""
		return int(self.lvm_entity()[10])

	def uuid(self):
		""" Return physical volume UUID

		:return: str
		"""
		return self.lvm_entity()[11]


class WVolumeGroup(WLVMInfo):
	""" Class represent a volume group
	"""

	@verify_type('paranoid', volume_group=(str, tuple, list, set), sudo=bool)
	@verify_value('paranoid', volume_group=lambda x: len(x) > 0 if isinstance(x, str) else True)
	def __init__(self, volume_group, sudo=False):
		""" Create new volume group descriptor

		:param volume_group: same as 'lvm_entity' in :meth:`.WLVMInfo.__init__`
		:param sudo: same as 'sudo' in :meth:`.WLVMInfo.__init__`
		"""
		WLVMInfo.__init__(self, 'vgdisplay', 17, volume_group, sudo=sudo)

	def all(self):
		""" Return every volume group in the system

		:return: tuple of WVolumeGroup
		"""
		return tuple([WVolumeGroup(x) for x in self.lvm_command().lvm_info()])

	def group_name(self):
		""" Return volume group name

		:return: str
		"""
		return self.lvm_entity()[0]

	def group_access(self):
		""" Return volume group access

		:return: str
		"""
		return self.lvm_entity()[1]

	def maximum_logical_volumes(self):
		""" Return maximum number of logical volumes (0 - for unlimited)

		:return: int
		"""
		return int(self.lvm_entity()[4])

	def logical_volumes(self):
		""" Return current number of logical volumes

		:return: int
		"""
		return int(self.lvm_entity()[5])

	def opened_logical_volumes(self):
		""" Return open count of all logical volumes in this volume group

		:return: int
		"""
		return int(self.lvm_entity()[6])

	def maximum_physical_volumes(self):
		""" Return maximum number of physical volumes (0 - for unlimited)

		:return: int
		"""
		return int(self.lvm_entity()[8])

	def physical_volumes(self):
		""" Return current number of physical volumes

		:return: int
		"""
		return int(self.lvm_entity()[9])

	def actual_volumes(self):
		""" Return actual number of physical volumes

		:return: int
		"""
		return int(self.lvm_entity()[10])

	def size(self):
		""" Return size of volume group in kilobytes

		:return: int
		"""
		return int(self.lvm_entity()[11])

	def extent_size(self):
		""" Return physical extent size in kilobytes

		:return: int
		"""
		return int(self.lvm_entity()[12])

	def total_extents(self):
		""" Return total number of physical extents for this volume group

		:return: int
		"""
		return int(self.lvm_entity()[13])

	def allocated_extents(self):
		""" Return allocated number of physical extents for this volume group

		:return: int
		"""
		return int(self.lvm_entity()[14])

	def free_extents(self):
		""" Return free number of physical extents for this volume group

		:return: int
		"""
		return int(self.lvm_entity()[15])

	def uuid(self):
		""" Return UUID of volume group

		:return: str
		"""
		return self.lvm_entity()[16]


class WLogicalVolume(WLVMInfo):
	""" Class represent a logical volume
	"""

	__lvm_snapshot_create_cmd_timeout__ = 3
	""" Timeout for snapshot creation command to complete
	"""

	__lvm_snapshot_remove_cmd_timeout__ = 3
	""" Timeout for snapshot removing command to complete
	"""

	__lvm_snapshot_check_cmd_timeout__ = 3
	""" Timeout for snapshot checking (getting parameters) command to complete
	"""

	__snapshot_maximum_allocation__ = 99.9
	""" Maximum space usage for snapshot, till that value snapshot is treated as valid
	"""

	@verify_type('paranoid', logical_volume=(str, tuple, list, set), sudo=bool)
	@verify_value('paranoid', logical_volume=lambda x: len(x) > 0 if isinstance(x, str) else True)
	def __init__(self, logical_volume, sudo=False):
		""" Create new logical volume descriptor

		:param logical_volume: same as 'lvm_entity' in :meth:`.WLVMInfo.__init__`
		:param sudo: same as 'sudo' in :meth:`.WLVMInfo.__init__`
		"""
		WLVMInfo.__init__(self, 'lvdisplay', 13, logical_volume, sudo=sudo)

	def all(self):
		""" Return every logical volume in the system

		:return: tuple of WLogicalVolume
		"""
		return tuple([WLogicalVolume(x) for x in self.lvm_command().lvm_info()])

	def volume_path(self):
		""" Return logical volume path

		:return: str
		"""
		return self.lvm_entity()[0]

	def volume_name(self):
		""" Return logical volume name

		:return: str
		"""
		return os.path.basename(self.volume_path())

	def volume_group_name(self):
		""" Return volume group name

		:return: str
		"""
		return self.lvm_entity()[1]

	def volume_group(self):
		""" Return volume group

		:return: WVolumeGroup
		"""
		return WVolumeGroup(self.volume_group_name(), sudo=self.lvm_command().sudo())

	def sectors_count(self):
		""" Return logical volume size in sectors

		:return: int
		"""
		return int(self.lvm_entity()[6])

	def extents_count(self):
		""" Return current logical extents associated to logical volume

		:return: int
		"""
		return int(self.lvm_entity()[7])

	def device_number(self):
		""" Return tuple of major and minor device number of logical volume

		:return: tuple of int
		"""
		return int(self.lvm_entity()[11]), int(self.lvm_entity()[12])

	def uuid(self):
		""" Return UUID of logical volume

		:return: str
		"""
		uuid_file = '/sys/block/%s/dm/uuid' % os.path.basename(os.path.realpath(self.volume_path()))
		lv_uuid = open(uuid_file).read().strip()
		if lv_uuid.startswith('LVM-') is True:
			return lv_uuid[4:]
		return lv_uuid

	@verify_type(snapshot_size=(int, float), snapshot_suffix=str)
	@verify_value(snapshot_size=lambda x: x > 0, snapshot_suffix=lambda x: len(x) > 0)
	def create_snapshot(self, snapshot_size, snapshot_suffix):
		""" Create snapshot for this logical volume.

		:param snapshot_size: size of newly created snapshot volume. This size is a fraction of the source \
		logical volume space (of this logical volume)
		:param snapshot_suffix: suffix for logical volume name (base part is the same as the original volume \
		name)

		:return: WLogicalVolume
		"""
		size_extent = math.ceil(self.extents_count() * snapshot_size)
		size_kb = self.volume_group().extent_size() * size_extent
		snapshot_name = self.volume_name() + snapshot_suffix

		lvcreate_cmd = ['sudo'] if self.lvm_command().sudo() is True else []

		lvcreate_cmd.extend([
			'lvcreate', '-L', '%iK' % size_kb, '-s', '-n', snapshot_name, '-p', 'r', self.volume_path()
		])
		subprocess.check_output(lvcreate_cmd, timeout=self.__class__.__lvm_snapshot_create_cmd_timeout__)
		return WLogicalVolume(self.volume_path() + snapshot_suffix, sudo=self.lvm_command().sudo())

	def remove_volume(self):
		""" Remove this volume

		:return: None
		"""
		lvremove_cmd = ['sudo'] if self.lvm_command().sudo() is True else []
		lvremove_cmd.extend(['lvremove', '-f', self.volume_path()])
		subprocess.check_output(lvremove_cmd, timeout=self.__class__.__lvm_snapshot_remove_cmd_timeout__)

	def snapshot_allocation(self):
		""" Return allocated size (fraction of total snapshot volume space). If this is not a snapshot volume,
		than RuntimeError exception is raised.

		:return: float
		"""
		check_cmd = ['lvs', self.volume_path(), '-o', 'snap_percent', '--noheadings']
		output = subprocess.check_output(check_cmd, timeout=self.__class__.__lvm_snapshot_check_cmd_timeout__)
		output = output.decode().strip()
		if len(output) == 0:
			raise RuntimeError('Unable to check general logical volume')
		return float(output.replace(',', '.', 1))

	def snapshot_corrupted(self):
		""" Check if this snapshot volume is corrupted or not

		:return: bool (True if corrupted, False - otherwise)
		"""
		return self.snapshot_allocation() > self.__class__.__snapshot_maximum_allocation__

	@classmethod
	@verify_type('paranoid', file_path=str, sudo=bool)
	@verify_value('paranoid', file_path=lambda x: len(x) > 0)
	def logical_volume(cls, file_path, sudo=False):
		""" Return logical volume that stores the given path

		:param file_path: target path to search
		:param sudo: same as 'sudo' in :meth:`.WLogicalVolume.__init__`
		:return: WLogicalVolume or None (if file path is outside current mount points)
		"""
		mp = WMountPoint.mount_point(file_path)
		if mp is not None:
			name_file = '/sys/block/%s/dm/name' % mp.device_name()
			if os.path.exists(name_file):
				lv_path = '/dev/mapper/%s' % open(name_file).read().strip()
				return WLogicalVolume(lv_path, sudo=sudo)



# -*- coding: utf-8 -*-

import pytest
import socket

from wasp_general.config import WConfig

from wasp_general.network.primitives import WIPV4SocketInfo
from wasp_general.network.beacon.transport import WBroadcastBeaconTransport, WMulticastBeaconTransport


class TestWBroadcastBeaconTransport:

	def test_target_socket(self):
		config = WConfig()
		config.add_section('wasp-general::network::beacon')
		config['wasp-general::network::beacon']['address'] = ''
		config['wasp-general::network::beacon']['port'] = ''

		t = WBroadcastBeaconTransport()
		pytest.raises(ValueError, t.target_socket, config)
		config['wasp-general::network::beacon']['address'] = 'site1.org'
		pytest.raises(ValueError, t.target_socket, config)
		config['wasp-general::network::beacon']['address'] = ''
		config['wasp-general::network::beacon']['port'] = '9999'
		pytest.raises(ValueError, t.target_socket, config)

		config['wasp-general::network::beacon']['address'] = 'site1.org'
		pytest.raises(ValueError, t.target_socket, config)

		config['wasp-general::network::beacon']['address'] = '4.4.4.4'
		assert(isinstance(t.target_socket(config), WIPV4SocketInfo) is True)
		assert(t.target_socket(config).pair() == ('4.4.4.4', 9999))

	def test_server_socket(self):
		config = WConfig()
		config.add_section('wasp-general::network::beacon')
		config['wasp-general::network::beacon']['bind_address'] = ''
		config['wasp-general::network::beacon']['port'] = '20202'

		t = WBroadcastBeaconTransport()
		s1 = t.server_socket(config)
		assert(isinstance(s1, socket.socket) is True)
		assert(s1.getsockname() == ('0.0.0.0', 20202))
		assert(s1 == t.server_socket(config))
		t.close_server_socket(config)

		config['wasp-general::network::beacon']['bind_address'] = '127.0.0.1'
		s2 = t.server_socket(config)
		assert(s2.getsockname() == ('127.0.0.1', 20202))
		assert(s2 != s1)
		t.close_server_socket(config)

	def test_client_socket(self):
		config = WConfig()
		config.add_section('wasp-general::network::beacon')
		config['wasp-general::network::beacon']['address'] = '127.0.0.1'
		config['wasp-general::network::beacon']['port'] = '20202'

		t = WBroadcastBeaconTransport()
		s1 = t.client_socket(config)
		assert(isinstance(s1, socket.socket) is True)
		assert(s1 == t.client_socket(config))
		t.close_client_socket(config)

		s2 = t.client_socket(config)
		assert(s2 != s1)
		t.close_client_socket(config)


class TestWMulticastBeaconTransport:

	def test_transport(self):
		config = WConfig()
		config.add_section('wasp-general::network::beacon')
		config['wasp-general::network::beacon']['bind_address'] = ''
		config['wasp-general::network::beacon']['address'] = '127.0.0.1'
		config['wasp-general::network::beacon']['port'] = '7070'
		transport = WMulticastBeaconTransport()

		pytest.raises(ValueError, transport.target_socket, config)
		config['wasp-general::network::beacon']['address'] = '239.200.1.2'
		assert(isinstance(transport.target_socket(config), WIPV4SocketInfo) is True)

		s1 = transport.server_socket(config)
		assert(isinstance(s1, socket.socket) is True)
		transport.close_server_socket(config)
		s2 = transport.server_socket(config)
		assert(isinstance(s2, socket.socket) is True)
		assert(s1 != s2)



# -*- coding: utf-8 -*-
# wasp_general/network/web/session.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: make code more useful
# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_subclass, verify_type
from wasp_general.network.primitives import WIPV4SocketInfo

from wasp_general.network.web.proto import WWebSessionProto, WWebResponseProto
from wasp_general.network.web.request import WWebRequest


class WWebSessionAdapter(WWebSessionProto):

	def client_address(self):
		return WIPV4SocketInfo(*self.accepted_socket().getpeername())

	def server_address(self):
		return WIPV4SocketInfo(*self.accepted_socket().getsockname())

	@abstractmethod
	def accepted_socket(self):
		raise NotImplementedError('This method is abstract')


class WWebSessionBase(WWebSessionAdapter, metaclass=ABCMeta):
	""" Basic :class:`.WWebSessionProto` implementation. This class clarifies prototype and appends several methods
	"""
	#TODO: Continue development!

	@verify_subclass(request_cls=WWebRequest)
	def __init__(self, request_cls=WWebRequest):
		""" Construct class

		:param request_cls: request class to use
		"""
		self.__request_cls = request_cls

	@verify_type(request_line=str)
	def read_request_line(self, request_line):
		""" Read HTTP-request line

		:param request_line: line to parse
			for HTTP/0.9 is GET <Request-URI>
			for HTTP/1.0 and 1.1 is <METHOD> <Request-URI> HTTP/<HTTP-Version>, where HTTP-Version is 1.0
			or 1.1.
			for HTTP/2: binary headers are used
		"""

		request = self.__request_cls.parse_request_line(self, request_line)

		protocol_version = self.protocol_version()
		if protocol_version == '0.9':
			if request.method() != 'GET':
				raise Exception('HTTP/0.9 standard violation')
		elif protocol_version == '1.0' or protocol_version == '1.1':
			pass
		elif protocol_version == '2':
			pass
		else:
			raise RuntimeError('Unsupported HTTP-protocol')



# -*- coding: utf-8 -*-

import pytest

from wasp_general.network.messenger.proto import WMessengerOnionLayerProto, WMessengerOnionSessionFlowProto
from wasp_general.network.messenger.proto import WMessengerOnionProto
from wasp_general.network.messenger.envelope import WMessengerTextEnvelope
from wasp_general.network.messenger.session import WMessengerOnionSession, WMessengerOnionSessionFlow

from wasp_general.network.messenger.layers import WMessengerOnionCoderLayerProto


def test_abstract():
	envelope = WMessengerTextEnvelope('')
	onion = TestWMessengerOnionCoderLayerBase.Onion()
	session_flow = WMessengerOnionSessionFlow.sequence_flow(WMessengerOnionSessionFlowProto.IteratorInfo('layer'))
	session = WMessengerOnionSession(onion, session_flow)

	pytest.raises(TypeError, WMessengerOnionCoderLayerProto)
	pytest.raises(NotImplementedError, WMessengerOnionCoderLayerProto.encode, None, envelope, session)
	pytest.raises(NotImplementedError, WMessengerOnionCoderLayerProto.decode, None, envelope, session)


class TestWMessengerOnionCoderLayerBase:

	class Onion(WMessengerOnionProto):
		def __init__(self):
			self.storage = {}

		def layers_names(self):
			return self.storage.keys()

		def layer(self, layer_name):
			return self.storage[layer_name]

	class Coder(WMessengerOnionCoderLayerProto):
		state = []

		def encode(self, envelope, session, **kwargs):
			self.state.append('encoded')
			return envelope

		def decode(self, envelope, session, **kwargs):
			self.state.append('decoded')
			return envelope

	def test(self):

		c = TestWMessengerOnionCoderLayerBase.Coder('layer1')
		assert(isinstance(c, WMessengerOnionCoderLayerProto) is True)
		assert(isinstance(c, WMessengerOnionLayerProto) is True)

		envelope = WMessengerTextEnvelope('')
		onion = TestWMessengerOnionCoderLayerBase.Onion()
		sf = WMessengerOnionSessionFlow.sequence_flow(WMessengerOnionSessionFlowProto.IteratorInfo('layer'))
		session = WMessengerOnionSession(onion, sf)
		pytest.raises(RuntimeError, c.process, envelope, session)

		assert(TestWMessengerOnionCoderLayerBase.Coder.state == [])
		c.process(envelope, session, mode=WMessengerOnionCoderLayerProto.Mode.encode)
		assert(TestWMessengerOnionCoderLayerBase.Coder.state == ['encoded'])

		c.process(envelope, session, mode=WMessengerOnionCoderLayerProto.Mode.decode)
		assert(TestWMessengerOnionCoderLayerBase.Coder.state == ['encoded', 'decoded'])

		pytest.raises(TypeError, c.process, envelope, session, mode=1)



# -*- coding: utf-8 -*-

import pytest

'''
from wasp_general.task.health import WTaskHealthSensor, WMeasurableTaskHealthSensor, WTaskHealth


def test_abstract_classes():
	pytest.raises(TypeError, WTaskHealthSensor)
	pytest.raises(NotImplementedError, WTaskHealthSensor.healthy, None)

	pytest.raises(TypeError, WMeasurableTaskHealthSensor)
	pytest.raises(NotImplementedError, WMeasurableTaskHealthSensor.value, None)
	pytest.raises(NotImplementedError, WMeasurableTaskHealthSensor.measurement_unit, None)

	pytest.raises(TypeError, WTaskHealth)


class TestWTaskHealthSensor:

	def test_sensor(self):

		class S(WTaskHealthSensor):
			def healthy(self):
				return True

		s1 = S('recommended hdd sensor', WTaskHealthSensor.WTaskSensorSeverity.recommended)
		assert(s1.name() == 'recommended hdd sensor')
		assert(s1.description() is None)
		assert(s1.severity() == WTaskHealthSensor.WTaskSensorSeverity.recommended)

		s2 = S('hdd critical sensor', WTaskHealthSensor.WTaskSensorSeverity.critical, description='bla-bla')
		assert (s2.name() == 'hdd critical sensor')
		assert (s2.description() == 'bla-bla')
		assert (s2.severity() == WTaskHealthSensor.WTaskSensorSeverity.critical)


class TestWMeasurableTaskHealthSensor:

	def test_sensor(self):

		class S(WMeasurableTaskHealthSensor):
			def value(self):
				return 1

		s1 = S('s1 sensor', WTaskHealthSensor.WTaskSensorSeverity.recommended, 10)
		assert(s1.name() == 's1 sensor')
		assert(s1.description() is None)
		assert(s1.severity() == WTaskHealthSensor.WTaskSensorSeverity.recommended)
		assert(s1.error_value() == 10)

		assert(s1.healthy() is True)
		s1.value = lambda: 11
		assert(s1.healthy() is False)

		s2 = S('s2 sensor', WTaskHealthSensor.WTaskSensorSeverity.recommended, 10, comparator=lambda x, y: x > y)
		assert(s2.healthy() is False)
		s2.value = lambda: 11
		assert(s2.healthy() is True)


class TestWTaskHealth:

	def test_healthy(self):

		class H(WTaskHealth):
			def start(self):
				pass

		h1 = H(decorate_stop=False)
		assert(h1.sensors() == [])
		assert(h1.healthy() is None)

		pytest.raises(TypeError, H, 1, decorate_stop=False)

		class S(WTaskHealthSensor):
			def healthy(self):
				return True

		s1 = S('sensor1', WTaskHealthSensor.WTaskSensorSeverity.recommended)
		s2 = S('sensor2', WTaskHealthSensor.WTaskSensorSeverity.important)
		h2 = H(s1, s2, decorate_stop=False)
		assert(h2.sensors() == ['sensor1', 'sensor2'] or h2.sensors() == ['sensor2', 'sensor1'])
		assert(h2.sensor('sensor1') == s1)
		assert(h2.sensor('sensor2') == s2)
		assert(h2.healthy() is None)

		s1.healthy = lambda: False
		assert(h2.healthy() == WTaskHealthSensor.WTaskSensorSeverity.recommended)
		s2.healthy = lambda: False
		assert(h2.healthy() == WTaskHealthSensor.WTaskSensorSeverity.important)

		s1.severity = lambda: WTaskHealthSensor.WTaskSensorSeverity.critical
		assert(h2.healthy() == WTaskHealthSensor.WTaskSensorSeverity.critical)
'''



# -*- coding: utf-8 -*-
# wasp_general/template_command.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from wasp_general.verify import verify_type
from wasp_general.command.result import WPlainCommandResult
from wasp_general.command.command import WCommand
from wasp_general.template import WTemplateRenderer, WTemplate


class WCommandResultTemplate(WPlainCommandResult, WTemplateRenderer):

	@verify_type('paranoid', template=WTemplate)
	def __init__(self, template, **command_env):
		WPlainCommandResult.__init__(self, '', **command_env)
		WTemplateRenderer.__init__(self, template)
		self.__template = template

	def __str__(self):
		return self.render()


class WTemplateResultCommand(WCommand):
	""" Simple command that returns "template-result", that can be "rendered" to str object later. In most time,
	this and derived classes return static template, which result is the same. The only requirement for execution
	of this command is to full command match. If :class:`.WTemplateResultCommand` was constructed with tokens like
	"foo", "bar" then only "foo bar" command will be matched.  (Note. For example, if :class:`.WCommand` object
	was constructed with tokens: "foo", "bar", then any of the following commands will be matched: "foo bar",
	"foo bar 1", "foo bar 2 3".)
	"""

	@verify_type('paranoid', command_tokens=str)
	@verify_type(template=WTemplate, template_context=(dict, None))
	def __init__(self, template, *command_tokens, template_context=None):
		""" Create new static template command

		:param template: template to use as result
		:param command_tokens: tokens that is used for command matching
		:param template_context: context that is used for rendering the template
		"""
		WCommand.__init__(self, *command_tokens)
		self.__template = template
		self.__template_context = template_context if template_context is not None else {}

	def template(self):
		""" Return template object (that will be used in result)

		:return: WTemplate
		"""
		return self.__template

	def template_context(self):
		""" Return context with which template will be rendered

		:return: dict
		"""
		return self.__template_context

	def result_template(self, *command_tokens, **command_env):
		""" Generate template result. command_tokens and command_env arguments are used for template
		detailing

		:param command_tokens: same as command_tokens in :meth:`.WCommandProto.match` and \
		:meth:`.WCommandProto.exec` methods (so as :meth:`.WCommand._exec`)
		:param command_env: same as command_env in :meth:`.WCommandProto.match` and \
		:meth:`.WCommandProto.exec` methods (so as :meth:`.WCommand._exec`)

		:return: WCommandResultTemplate
		"""
		result = WCommandResultTemplate(self.template())
		result.update_context(**self.template_context())
		return result

	def match(self, *command_tokens, **command_env):
		""" same as :meth:`.WCommand.meth` method, but checks for extra arguments also (no extra arguments is
		allowed)
		"""
		if command_tokens == self.command():
			return True
		return False

	def _exec(self, *command_tokens, **command_env):
		""" :meth:`.WCommand._exec` implementation
		"""
		return self.result_template(*command_tokens, **command_env)



# -*- coding: utf-8 -*-
# wasp_general/crypto/hmac.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from Crypto.Hash.HMAC import HMAC
import re

from wasp_general.verify import verify_type, verify_value
from wasp_general.crypto.hash import WHash


class WHMAC:
	""" Class that wraps PyCrypto HMAC implementation

	see also https://en.wikipedia.org/wiki/Hash-based_message_authentication_code
	"""

	__default_generator_name__ = 'SHA512'
	""" Default hash generator name for HMAC
	"""

	__hmac_name_re__ = re.compile('HMAC[\-_]([a-zA-Z0-9]+)')
	""" Regular expression that selects hash generator name from HMAC name
	"""

	@verify_type(digest_generator=(str, None))
	def __init__(self, digest_generator_name=None):
		""" Create new "code-authenticator"

		:param digest_generator_name: name of hash function
		"""
		if digest_generator_name is None:
			digest_generator_name = WHMAC.__default_generator_name__
		if digest_generator_name not in WHash.available_generators():
			raise ValueError('Unknown hash generator: "%s"' % digest_generator_name)
		self.__digest_generator = WHash.generator(digest_generator_name)

	def digest_generator(self):
		""" Return hash-generator

		:return: PyCrypto class
		"""
		return self.__digest_generator

	@verify_type(key=bytes, message=(bytes, None), digest_generator=(str, None))
	def hash(self, key, message=None):
		""" Return digest of the given message and key

		:param key: secret HMAC key
		:param message: code (message) to authenticate

		:return: bytes
		"""
		generator = self.digest_generator()
		return HMAC(key, msg=message, digestmod=generator().pycrypto()).digest()

	@classmethod
	@verify_type(name=str)
	@verify_value(name=lambda x: WHMAC.__hmac_name_re__.match(x) is not None)
	def hmac(cls, name):
		""" Return new WHMAC object by the given algorithm name like 'HMAC-SHA256' or 'HMAC_SHA1'

		:param name: name of HMAC algorithm

		:return: WHMAC
		"""
		return WHMAC(cls.__hmac_name_re__.search(name).group(1))



# -*- coding: utf-8 -*-
# wasp_general/config.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from configparser import ConfigParser, NoOptionError, NoSectionError
import os

from wasp_general.verify import verify_type, verify_value


class WConfig(ConfigParser):
	""" Improved ConfigParser. Has single method to merge config data (see :meth:`.WConfig.merge` method) and
	method, that split coma-separated option value (see :meth:`.WConfig.split_option` method).
	"""

	@verify_type(section=str, option=str)
	def split_option(self, section, option):
		""" Return list of strings that are made by splitting coma-separated option value. Method returns
		empty list if option value is empty string

		:param section: option section name
		:param option: option name
		:return: list of strings
		"""
		value = self[section][option].strip()
		if value == "":
			return []
		return [x.strip() for x in (value.split(","))]

	@verify_type(config=(str, ConfigParser))
	@verify_value(config=lambda x: isinstance(x, ConfigParser) or os.path.isfile(x))
	def merge(self, config):
		""" Load configuration from given configuration.

		:param config: config to load. If config is a string type, then it's treated as .ini filename
		:return: None
		"""
		if isinstance(config, ConfigParser) is True:
			self.update(config)
		elif isinstance(config, str):
			self.read(config)

	@verify_type(config=ConfigParser, section_to=str, section_from=(str, None))
	def merge_section(self, config, section_to, section_from=None):
		""" Load configuration section from other configuration. If specified section doesn't exist in current
		configuration, then it will be added automatically.

		:param config: source configuration
		:param section_to: destination section name
		:param section_from: source section name (if it is None, then section_to is used as source section name)
		:return: None
		"""
		section_from = section_from if section_from is not None else section_to
		if section_from not in config.sections():
			raise ValueError('There is no such section "%s" in config' % section_from)

		if section_to not in self.sections():
			self.add_section(section_to)

		for option in config[section_from].keys():
			self.set(section_to, option, config[section_from][option])

	def select_options(self, section, options_prefix=None):
		""" Select options from section

		:param section: target section
		:param options_prefix: prefix of options that should be selected

		:return: WConfigSelection
		"""
		return WConfigSelection(self, section, options_prefix=options_prefix)


class WConfigSelection:
	""" Represent subset of options in a single configuration section. Has methods __str__, __int__, __float__,
	__bool__ that return option value with option name equals to options prefix.
	"""

	@verify_type(config=WConfig, section=str, options_prefix=(str, None))
	@verify_value(section=lambda x: len(x) > 0)
	def __init__(self, config, section, options_prefix=None):
		""" Create a configuration selection

		:param config: source configuration
		:param section: source section name
		:param options_prefix: name prefix of options that should be selected. If it is not specified, then \
		it is treated as empty string
		"""
		self.__config = config
		self.__section = section
		self.__options_prefix = options_prefix if options_prefix is not None else ''

		if self.__config.has_section(section) is False:
			raise NoSectionError('Invalid section "%s" was specified' % section)

	def config(self):
		""" Return source configuration

		:return: WConfig
		"""
		return self.__config

	def section(self):
		""" Return source section name

		:return: str
		"""
		return self.__section

	def option_prefix(self):
		""" Return name prefix of options that may be selected

		:return: str
		"""
		return self.__options_prefix

	def __option(self):
		""" Check and return option from section from configuration. Option name is equal to option prefix

		:return: tuple of section name and option prefix
		"""
		section = self.section()
		option = self.option_prefix()
		if self.config().has_option(section, option) is False:
			raise NoOptionError(option, section)
		return section, option

	def __str__(self):
		""" Return string value of this option

		:return: str
		"""
		section, option = self.__option()
		return self.config()[section][option]

	def __int__(self):
		""" Return integer value of this option

		:return: int
		"""
		section, option = self.__option()
		return self.config().getint(section, option)

	def __float__(self):
		""" Return float value of this option

		:return: float
		"""
		section, option = self.__option()
		return self.config().getfloat(section, option)

	def __bool__(self):
		""" Return boolean value of this option

		:return: bool
		"""
		section, option = self.__option()
		return self.config().getboolean(section, option)

	@verify_type(option_prefix=str)
	@verify_value(option_prefix=lambda x: len(x) > 0)
	def select_options(self, options_prefix):
		""" Select options from this selection, that are started with the specified prefix

		:param options_prefix: name prefix of options that should be selected
		:return: WConfigSelection
		"""
		return WConfigSelection(
			self.config(), self.section(), self.option_prefix() + options_prefix
		)

	@verify_type('paranoid', item=str)
	@verify_value('paranoid', item=lambda x: len(x) > 0)
	def __getitem__(self, item):
		""" Alias for :meth:`.WConfigSelection.select_options` method

		:param item:
		:return:
		"""
		return self.select_options(item)

	@verify_type(option_name=(str, None))
	def has_option(self, option_name=None):
		""" Check whether configuration selection has the specified option.

		:param option_name: option name to check. If no option is specified, then check is made for this option

		:return: bool
		"""
		if option_name is None:
			option_name = ''
		return self.config().has_option(self.section(), self.option_prefix() + option_name)



# -*- coding: utf-8 -*-
# wasp_general/crypto/kdf.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from Crypto.Protocol.KDF import PBKDF2

from wasp_general.verify import verify_type, verify_value
from wasp_general.crypto.hmac import WHMAC
from wasp_general.crypto.random import random_bytes


class WPBKDF2:
	""" Wrapper for PyCrypto PBKDF2 implementation with NIST recommendation and HMAC is used as pseudorandom
	function

	NIST recommendation can be read here:
	http://nvlpubs.nist.gov/nistpubs/Legacy/SP/nistspecialpublication800-132.pdf (Recommendation for Password-Based
	Key Derivation)
	"""

	__minimum_key_length__ = 20
	""" Minimum key length is specified at Appendix A, Section A.1 of Recommendation for Password-Based
	Key Derivation by NIST
	"""

	__minimum_salt_length__ = 16
	""" Minimum salt length is specified at Section 5.1 of Recommendation for Password-Based Key Derivation by NIST
	"""
	__default_salt_length__ = 64
	""" Salt length value used by default
	"""

	__default_digest_generator_name__ = 'SHA256'
	""" Hash-generator that is used by default
	"""

	__minimum_iterations_count__ = 1000
	""" Minimum iteration Count is specified at Section 5.2 of Recommendation for Password-Based Key Derivation by
	NIST
	"""
	__default_iterations_count__ = 1000
	""" The iteration count used by default
	"""

	__default_derived_key_length__ = 16
	""" Length of derived key used by default
	"""

	@verify_type(key=(str, bytes), salt=(bytes, None), derived_key_length=(int, None))
	@verify_type(iterations_count=(int, None), hmac=(WHMAC, None))
	@verify_value(key=lambda x: x is None or len(x) >= WPBKDF2.__minimum_key_length__)
	@verify_value(salt=lambda x: x is None or len(x) >= WPBKDF2.__minimum_salt_length__)
	@verify_value(iterations_count=lambda x: x is None or x >= WPBKDF2.__minimum_iterations_count__)
	def __init__(self, key, salt=None, derived_key_length=None, iterations_count=None, hmac=None):
		""" Generate new key (derived key) with PBKDF2 algorithm

		:param key: password
		:param salt: salt to use (if no salt was specified, then it will be generated automatically)
		:param derived_key_length: length of byte-sequence to generate
		:param iterations_count: iteration count
		:param hmac: WHMAC object to use with PBKDF2
		"""
		self.__salt = salt if salt is not None else self.generate_salt()
		if derived_key_length is None:
			derived_key_length = self.__default_derived_key_length__
		if iterations_count is None:
			iterations_count = self.__default_iterations_count__
		if hmac is None:
			hmac = WHMAC(self.__default_digest_generator_name__)

		self.__derived_key = PBKDF2(
			key, self.__salt, dkLen=derived_key_length, count=iterations_count,
			prf=hmac.hash
		)

	def salt(self):
		""" Return salt value (that was given in constructor or created automatically)

		:return: bytes
		"""
		return self.__salt

	def derived_key(self):
		""" Return derived key

		:return: bytes
		"""
		return self.__derived_key

	@classmethod
	@verify_type(length=(int, None))
	@verify_value(length=lambda x: x is None or x >= WPBKDF2.__minimum_salt_length__)
	def generate_salt(cls, length=None):
		""" Generate salt that can be used by this object

		:param length: target salt length

		:return: bytes
		"""
		if length is None:
			length = cls.__default_salt_length__
		return random_bytes(length)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.types.binarray import WBinArray
from wasp_general.types.bytearray import WFixedSizeByteArray


class TestWFixedSizeByteArray:

	def test_init(self):
		pytest.raises(ValueError, WFixedSizeByteArray, 0, WBinArray(1))
		pytest.raises(ValueError, WFixedSizeByteArray, 0, WBinArray(300))
		pytest.raises(OverflowError, WFixedSizeByteArray, 1, b'12')
		assert(isinstance(WFixedSizeByteArray(0), WFixedSizeByteArray) is True)
		assert(isinstance(WFixedSizeByteArray(1), WFixedSizeByteArray) is True)
		assert(isinstance(WFixedSizeByteArray(1, WBinArray(0)), WFixedSizeByteArray) is True)
		assert(isinstance(WFixedSizeByteArray(1, 0), WFixedSizeByteArray) is True)

	def test_bin_array(self):
		assert(isinstance(WFixedSizeByteArray(1, 1).bin_array(), list))
		assert(len((WFixedSizeByteArray(2, 1).bin_array())) == 2)
		assert(int((WFixedSizeByteArray(2, 1).bin_array()[0])) == 0)
		assert(int((WFixedSizeByteArray(2, 1).bin_array()[1])) == 1)

	def test_bin_value(self):
		assert(isinstance(WFixedSizeByteArray(1, 1).bin_value(), WBinArray))
		assert(str(WFixedSizeByteArray(1, 1).bin_value()) == '00000001')
		assert(str(WFixedSizeByteArray(2, 1).bin_value()) == '0000000000000001')

	def test_resize(self):
		a = WFixedSizeByteArray(0)
		pytest.raises(IndexError, a.__getitem__, 0)
		a.resize(1)
		assert(int(a[0]) == 0)
		pytest.raises(ValueError, a.resize, 0)

	def test_len(self):
		assert(len(WFixedSizeByteArray(0)) == 0)
		assert(len(WFixedSizeByteArray(1)) == 1)

	def test_str(self):
		assert(str(WFixedSizeByteArray(2, 1)) == "['00000000', '00000001']")

	def test_swipe(self):
		assert(str(WFixedSizeByteArray(2, 1).swipe()) == "['00000001', '00000000']")

	def test_getitem(self):
		assert(int(WFixedSizeByteArray(2, 256)[0]) == 1)
		assert(int(WFixedSizeByteArray(2, 256)[1]) == 0)
		assert(isinstance(WFixedSizeByteArray(3, 65536)[:2], list))
		assert(len(WFixedSizeByteArray(3, 65536)[:2]) == 2)
		assert(int(WBinArray.join(*WFixedSizeByteArray(3, 65536)[:2])) == 256)
		assert(int(WBinArray.join(*WFixedSizeByteArray(3, 65536)[1:])) == 0)

	def test_setitem(self):
		a = WFixedSizeByteArray(1)
		pytest.raises(IndexError, a.__setitem__, 2, 5)
		a[0] = WBinArray(1)
		assert(int(a[0]) == 1)
		a[0] = WBinArray(1, 8)
		assert(int(a[0]) == 1)
		a[0] = 1
		assert(int(a[0]) == 1)
		pytest.raises(ValueError, a.__setitem__, 0, 257)
		pytest.raises(ValueError, a.__setitem__, 0, WBinArray(257))
		a[0] = WBinArray(254, 9)
		assert(int(a[0]) == 254)

	def test_bytes(self):
		a = WFixedSizeByteArray(3)
		a[0] = 97
		a[1] = 115
		a[2] = 100
		assert(bytes(a) == b'asd')




from tempfile import mktemp
import os
import pytest


@pytest.fixture
def temp_file(request):
	filename = mktemp('pytest-wasp-general-')

	def fin():
		if os.path.exists(filename):
			os.unlink(filename)
	request.addfinalizer(fin)
	return filename



# -*- coding: utf-8 -*-

import pytest

from wasp_general.command.command import WCommandProto, WCommand, WCommandPrioritizedSelector
from wasp_general.command.command import WCommandSet
from wasp_general.command.result import WPlainCommandResult
from wasp_general.composer import WComposerProto

from wasp_general.command.context import WContextProto, WContext, WCommandContextAdapter, WCommandContext
from wasp_general.command.context import WContextComposer


def test_abstract():
	pytest.raises(TypeError, WContextProto)
	pytest.raises(NotImplementedError, WContextProto.context_name, None)
	pytest.raises(NotImplementedError, WContextProto.context_value, None)
	pytest.raises(NotImplementedError, WContextProto.linked_context, None)

	pytest.raises(TypeError, WCommandContextAdapter)
	pytest.raises(NotImplementedError, WCommandContextAdapter.adapt, None)


class TestWContextProto:

	class Context(WContextProto):

		def context_name(self):
			return 'context1'

		def context_value(self):
			return None

		def linked_context(self):
			return None

	def test(self):
		c1 = TestWContextProto.Context()
		assert(len(c1) == 1)

		c2 = TestWContextProto.Context()
		c2.linked_context = lambda: c1
		assert(len(c1) == 1)
		assert(len(c2) == 2)

		c3 = TestWContextProto.Context()
		c3.linked_context = lambda: c2
		assert(len(c1) == 1)
		assert(len(c2) == 2)
		assert(len(c3) == 3)


class TestWContext:

	def test(self):
		c_r1 = WContext('context1')
		assert(isinstance(c_r1, WContext) is True)
		assert(isinstance(c_r1, WContextProto) is True)
		assert(c_r1 != 1)

		assert(c_r1.context_name() == 'context1')
		assert(c_r1.context_value() is None)
		assert(c_r1.linked_context() is None)

		c_r2 = WContext('context2', 'context-value', c_r1)
		assert(c_r2.context_name() == 'context2')
		assert(c_r2.context_value() == 'context-value')
		assert(c_r2.linked_context() == c_r1)
		assert(c_r2 != c_r1)

		c_r3 = WContext('context1', None, c_r2)
		assert(c_r1 != c_r3)
		assert(c_r3 != c_r1)

		assert(WContext.export_context(c_r2) == (('context1', None), ('context2', 'context-value')))
		assert(WContext.export_context(None) is None)


class TestWContextComposer:

	def test(self):
		context1 = WContext('context1')
		context2 = WContext('context2', 'context-value', context1)

		composer = WContextComposer()
		assert(isinstance(composer, WContextComposer) is True)
		assert(isinstance(composer, WComposerProto) is True)

		decomposed_context = composer.decompose(context2)
		composed_context = composer.compose(decomposed_context)
		assert(composed_context == context2)


class TestWCommandContextAdapter:

	class Adapter(WCommandContextAdapter):

		def adapt(self, *command_tokens, request_context=None):
			if request_context is None:
				return command_tokens

			result = [request_context.context_name()]
			result.extend(command_tokens)
			return tuple(result)

	def test(self):
		spec = WContext.specification('context1', 'context2')
		a = TestWCommandContextAdapter.Adapter(spec)
		assert(isinstance(a, WCommandContextAdapter) is True)
		assert(a.specification() == spec)

		c_r1 = WContext('context1')
		c_r2 = WContext('context2', 'context-value', c_r1)

		assert(a.adapt('hello', 'world', request_context=c_r1) == ('context1', 'hello', 'world'))
		assert(a.adapt('hello', 'world', request_context=c_r2) == ('context2', 'hello', 'world'))

		assert(TestWCommandContextAdapter.Adapter(WContext.specification()).match() is True)
		assert(a.match() is False)
		assert(a.match(c_r1) is False)
		assert(a.match(c_r2) is True)


class TestWCommandContext:

	class Command(WCommand):

		def _exec(self, *command_tokens, **kwargs):
			return WPlainCommandResult('OK')

	class Adapter(WCommandContextAdapter):
		def adapt(self, *command_tokens, command_context=None):
			if command_context is None:
				return command_tokens

			result = [command_context.context_value()]
			result.extend(command_tokens)
			return tuple(result)

	def test(self):
		base_command = TestWCommandContext.Command('hello', 'world')
		adapter = TestWCommandContext.Adapter(WContext.specification('context'))
		command = WCommandContext(base_command, adapter)

		assert(isinstance(command, WCommandContext) is True)
		assert(isinstance(command, WCommandProto) is True)
		assert(command.original_command() == base_command)
		assert(command.adapter() == adapter)

		adapter2 = TestWCommandContext.Adapter(WContext.specification())
		command2 = WCommandContext(base_command, adapter2)
		assert(command2.match('hello', 'world', 'yeah') is True)
		assert(command.match('hello', 'world', 'yeah') is False)
		command_context = WContext('context', 'hello')
		assert(command.match('world', command_context=command_context) is True)

		assert(str(command2.exec('hello', 'world')) == 'OK')
		assert(str(command.exec('world', command_context=command_context)) == 'OK')
		pytest.raises(RuntimeError, command.exec, 'hello', 'world')
